from unittest import IsolatedAsyncioTestCase

from ....protocols.trustping.v1_0.messages.ping import Ping
from ....utils.testing import create_test_profile
from ....wallet.base import BaseWallet
from ....wallet.key_type import ED25519
from ..signature_decorator import SignatureDecorator

TEST_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"


class TestSignatureDecorator(IsolatedAsyncioTestCase):
    async def test_init(self):
        decorator = SignatureDecorator()
        assert decorator.signature_type is None
        assert decorator.signature is None
        assert decorator.sig_data is None
        assert decorator.signer is None
        assert "SignatureDecorator" in str(decorator)

    async def test_serialize_load(self):
        TEST_SIG = "IkJvYiI="
        TEST_SIG_DATA = "MTIzNDU2Nzg5MCJCb2Ii"

        decorator = SignatureDecorator(
            signature_type=SignatureDecorator.TYPE_ED25519SHA512,
            signature=TEST_SIG,
            sig_data=TEST_SIG_DATA,
            signer=TEST_VERKEY,
        )

        dumped = decorator.serialize()
        loaded = SignatureDecorator.deserialize(dumped)

        assert loaded.signature_type == SignatureDecorator.TYPE_ED25519SHA512
        assert loaded.signature == TEST_SIG
        assert loaded.sig_data == TEST_SIG_DATA
        assert loaded.signer == TEST_VERKEY

    async def test_create_decode_verify(self):
        TEST_MESSAGE = "Hello world"
        TEST_TIMESTAMP = 1234567890

        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            key_info = await wallet.create_signing_key(ED25519)

            deco = await SignatureDecorator.create(
                Ping(), key_info.verkey, wallet, timestamp=None
            )
            assert deco

            deco = await SignatureDecorator.create(
                TEST_MESSAGE, key_info.verkey, wallet, TEST_TIMESTAMP
            )

            (msg, timestamp) = deco.decode()
            assert msg == TEST_MESSAGE
            assert timestamp == TEST_TIMESTAMP

            await deco.verify(wallet)
            deco.signature_type = "unsupported-sig-type"
            assert not await deco.verify(wallet)



"""Module that contains singleton classes for wallet operations."""


class IsAnonCredsSingleton:
    """Singleton class used as cache for anoncreds wallet-type queries."""

    instance = None
    wallets = set()

    def __new__(cls, *args, **kwargs):
        """Create a new instance of the class."""
        if cls.instance is None:
            cls.instance = super().__new__(cls)
        return cls.instance

    def set_wallet(self, wallet: str):
        """Set a wallet name."""
        self.wallets.add(wallet)

    def remove_wallet(self, wallet: str):
        """Remove a wallet name."""
        self.wallets.discard(wallet)


class UpgradeInProgressSingleton:
    """Singleton class used as cache for upgrade in progress."""

    instance = None
    wallets = set()

    def __new__(cls, *args, **kwargs):
        """Create a new instance of the class."""
        if cls.instance is None:
            cls.instance = super().__new__(cls)
        return cls.instance

    def set_wallet(self, wallet: str):
        """Set a wallet name."""
        self.wallets.add(wallet)

    def remove_wallet(self, wallet: str):
        """Remove a wallet name."""
        self.wallets.discard(wallet)



"""Utilities to deal with indy."""

from typing import Mapping, Optional

from marshmallow import (
    EXCLUDE,
    Schema,
    ValidationError,
    fields,
    validate,
    validates_schema,
)

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.models.openapi import OpenAPISchema
from ...messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INT_EPOCH_EXAMPLE,
    INT_EPOCH_VALIDATE,
    MAJOR_MINOR_VERSION_EXAMPLE,
    MAJOR_MINOR_VERSION_VALIDATE,
    NUM_STR_NATURAL_EXAMPLE,
    NUM_STR_NATURAL_VALIDATE,
    PREDICATE_EXAMPLE,
    PREDICATE_VALIDATE,
)


class IndyProofReqAttrSpecSchema(OpenAPISchema):
    """Schema for attribute specification in indy proof request."""

    name = fields.Str(
        required=False,
        metadata={"example": "favouriteDrink", "description": "Attribute name"},
    )
    names = fields.List(
        fields.Str(metadata={"example": "age"}),
        required=False,
        metadata={"description": "Attribute name group"},
    )
    restrictions = fields.List(
        fields.Dict(
            keys=fields.Str(
                validate=validate.Regexp(
                    "^schema_id|schema_issuer_did|schema_name|schema_version|issuer_did|"
                    "cred_def_id|attr::.+::value$"
                ),
                metadata={"example": "cred_def_id"},
            ),
            values=fields.Str(metadata={"example": INDY_CRED_DEF_ID_EXAMPLE}),
        ),
        required=False,
        metadata={
            "description": (
                "If present, credential must satisfy one of given restrictions: specify"
                " schema_id, schema_issuer_did, schema_name, schema_version,"
                " issuer_did, cred_def_id, and/or attr::<attribute-name>::value where"
                " <attribute-name> represents a credential attribute name"
            )
        },
    )
    non_revoked = fields.Nested(
        Schema.from_dict(
            {
                "from": fields.Int(
                    required=False,
                    validate=INT_EPOCH_VALIDATE,
                    metadata={
                        "description": (
                            "Earliest time of interest in non-revocation interval"
                        ),
                        "strict": True,
                        "example": INT_EPOCH_EXAMPLE,
                    },
                ),
                "to": fields.Int(
                    required=False,
                    validate=INT_EPOCH_VALIDATE,
                    metadata={
                        "description": (
                            "Latest time of interest in non-revocation interval"
                        ),
                        "strict": True,
                        "example": INT_EPOCH_EXAMPLE,
                    },
                ),
            },
            name="IndyProofReqAttrSpecNonRevokedSchema",
        ),
        allow_none=True,
        required=False,
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Data must have exactly one of name or names; if names then restrictions are
        mandatory.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: if data has both or neither of name and names

        """
        if ("name" in data) == ("names" in data):
            raise ValidationError(
                "Attribute specification must have either name or names but not both"
            )
        restrictions = data.get("restrictions")
        if ("names" in data) and (not restrictions or all(not r for r in restrictions)):
            raise ValidationError(
                "Attribute specification on 'names' must have non-empty restrictions"
            )


class IndyProofReqPredSpecSchema(OpenAPISchema):
    """Schema for predicate specification in indy proof request."""

    name = fields.Str(
        required=True, metadata={"example": "index", "description": "Attribute name"}
    )
    p_type = fields.Str(
        required=True,
        validate=PREDICATE_VALIDATE,
        metadata={
            "description": "Predicate type ('<', '<=', '>=', or '>')",
            "example": PREDICATE_EXAMPLE,
        },
    )
    p_value = fields.Int(
        required=True, metadata={"description": "Threshold value", "strict": True}
    )
    restrictions = fields.List(
        fields.Dict(
            keys=fields.Str(
                validate=validate.Regexp(
                    "^schema_id|schema_issuer_did|schema_name|schema_version|issuer_did|"
                    "cred_def_id|attr::.+::value$"
                ),
                metadata={"example": "cred_def_id"},
            ),
            values=fields.Str(metadata={"example": INDY_CRED_DEF_ID_EXAMPLE}),
        ),
        required=False,
        metadata={
            "description": (
                "If present, credential must satisfy one of given restrictions: specify"
                " schema_id, schema_issuer_did, schema_name, schema_version,"
                " issuer_did, cred_def_id, and/or attr::<attribute-name>::value where"
                " <attribute-name> represents a credential attribute name"
            )
        },
    )
    non_revoked = fields.Nested(
        Schema.from_dict(
            {
                "from": fields.Int(
                    required=False,
                    validate=INT_EPOCH_VALIDATE,
                    metadata={
                        "description": (
                            "Earliest time of interest in non-revocation interval"
                        ),
                        "strict": True,
                        "example": INT_EPOCH_EXAMPLE,
                    },
                ),
                "to": fields.Int(
                    required=False,
                    validate=INT_EPOCH_VALIDATE,
                    metadata={
                        "description": (
                            "Latest time of interest in non-revocation interval"
                        ),
                        "strict": True,
                        "example": INT_EPOCH_EXAMPLE,
                    },
                ),
            },
            name="IndyProofReqPredSpecNonRevokedSchema",
        ),
        allow_none=True,
        required=False,
    )


class IndyProofRequest(BaseModel):
    """Indy proof request."""

    class Meta:
        """Indy proof request metadata."""

        schema_class = "IndyProofRequestSchema"

    def __init__(
        self,
        nonce: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
        requested_attributes: Optional[Mapping] = None,
        requested_predicates: Optional[Mapping] = None,
        non_revoked: Optional[Mapping] = None,
        **kwargs,
    ):
        """Initialize indy cred abstract object.

        Args:
            nonce (str): The nonce value.
            name (str): The name of the proof request.
            version (str): The version of the proof request.
            requested_attributes (Mapping): A mapping of attribute names to attribute
                constraints.
            requested_predicates (Mapping): A mapping of predicate names to predicate
                constraints.
            non_revoked (Mapping): A mapping of non-revocation timestamps.
            kwargs: Keyword arguments for BaseModel

        """
        super().__init__(**kwargs)
        self.nonce = nonce
        self.name = name
        self.version = version
        self.requested_attributes = requested_attributes
        self.requested_predicates = requested_predicates
        self.non_revoked = non_revoked


class IndyProofRequestSchema(BaseModelSchema):
    """Schema for indy proof request."""

    class Meta:
        """Indy proof request schema metadata."""

        model_class = IndyProofRequest
        unknown = EXCLUDE

    nonce = fields.Str(
        required=False,
        validate=NUM_STR_NATURAL_VALIDATE,
        metadata={"description": "Nonce", "example": NUM_STR_NATURAL_EXAMPLE},
    )
    name = fields.Str(
        required=False,
        dump_default="Proof request",
        metadata={"description": "Proof request name", "example": "Proof request"},
    )
    version = fields.Str(
        required=False,
        dump_default="1.0",
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Proof request version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    requested_attributes = fields.Dict(
        required=True,
        keys=fields.Str(
            metadata={"decription": "Attribute referent", "example": "0_legalname_uuid"}
        ),
        values=fields.Nested(IndyProofReqAttrSpecSchema()),
        metadata={"description": "Requested attribute specifications of proof request"},
    )
    requested_predicates = fields.Dict(
        required=True,
        keys=fields.Str(
            metadata={"description": "Predicate referent", "example": "0_age_GE_uuid"}
        ),
        values=fields.Nested(IndyProofReqPredSpecSchema()),
        metadata={"description": "Requested predicate specifications of proof request"},
    )
    non_revoked = fields.Nested(
        Schema.from_dict(
            {
                "from": fields.Int(
                    required=False,
                    validate=INT_EPOCH_VALIDATE,
                    metadata={
                        "description": (
                            "Earliest time of interest in non-revocation interval"
                        ),
                        "strict": True,
                        "example": INT_EPOCH_EXAMPLE,
                    },
                ),
                "to": fields.Int(
                    required=False,
                    validate=INT_EPOCH_VALIDATE,
                    metadata={
                        "description": (
                            "Latest time of interest in non-revocation interval"
                        ),
                        "strict": True,
                        "example": INT_EPOCH_EXAMPLE,
                    },
                ),
            },
            name="IndyProofRequestNonRevokedSchema",
        ),
        allow_none=True,
        required=False,
    )



import json
from unittest import TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import FORWARD, PROTOCOL_PACKAGE
from ..forward import Forward, ForwardSchema


class TestForward(TestCase):
    to = "to"
    msg = {"msg": "body"}

    def setUp(self):
        self.message = Forward(to=self.to, msg=self.msg)

    def test_init(self):
        assert self.message.to == self.to

    def test_type(self):
        assert self.message._type == DIDCommPrefix.qualify_current(FORWARD)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.forward.ForwardSchema.load")
    def test_deserialize(self, message_schema_load):
        obj = {"obj": "obj"}

        message = Forward.deserialize(obj)
        message_schema_load.assert_called_once_with(obj)

        assert message is message_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.forward.ForwardSchema.dump")
    def test_serialize(self, message_schema_dump):
        message_dict = self.message.serialize()
        message_schema_dump.assert_called_once_with(self.message)

        assert message_dict is message_schema_dump.return_value


class TestForwardSchema(TestCase):
    def test_make_model(self):
        message = Forward(to="to", msg={"some": "msg"})
        data = message.serialize()
        model_instance = Forward.deserialize(data)
        assert isinstance(model_instance, Forward)

    def test_make_model_str(self):
        MSG = {"some": "msg"}
        message = Forward(to="to", msg=json.dumps(MSG))
        data = message.serialize()
        model_instance = Forward.deserialize(data)
        assert isinstance(model_instance, Forward)

        assert {"msg": MSG} == ForwardSchema().handle_str_message(
            data={"msg": json.dumps(MSG)}
        )



import asyncio
from time import time
from unittest import IsolatedAsyncioTestCase

from anoncreds import CredentialDefinitionPrivate, KeyCorrectnessProof
from aries_askar import AskarError

from ...anoncreds.issuer import CATEGORY_CRED_DEF_PRIVATE
from ...askar.profile import AskarProfileSession
from ...cache.base import BaseCache
from ...indy.credx.issuer import CATEGORY_CRED_DEF_KEY_PROOF
from ...ledger.multiple_ledger.ledger_requests_executor import IndyLedgerRequestsExecutor
from ...messaging.credential_definitions.util import CRED_DEF_SENT_RECORD_TYPE
from ...messaging.schemas.util import SCHEMA_SENT_RECORD_TYPE
from ...multitenant.base import BaseMultitenantManager
from ...multitenant.manager import MultitenantManager
from ...storage.base import BaseStorage
from ...storage.record import StorageRecord
from ...storage.type import (
    RECORD_TYPE_ACAPY_STORAGE_TYPE,
    RECORD_TYPE_ACAPY_UPGRADING,
    STORAGE_TYPE_VALUE_ANONCREDS,
)
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import anoncreds_upgrade


class TestAnonCredsUpgrade(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile(settings={"wallet.type": "askar"})
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, mock.MagicMock(MultitenantManager, autospec=True)
        )
        self.profile.context.injector.bind_instance(
            BaseCache, mock.MagicMock(BaseCache, autospec=True)
        )

    @mock.patch.object(
        CredentialDefinitionPrivate,
        "load",
        mock.MagicMock(
            return_value=mock.MagicMock(to_json_buffer=mock.MagicMock(return_value=b"{}"))
        ),
    )
    @mock.patch.object(
        KeyCorrectnessProof,
        "load",
        mock.MagicMock(
            return_value=mock.MagicMock(to_json_buffer=mock.MagicMock(return_value=b"{}"))
        ),
    )
    async def test_convert_records_to_anoncreds(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)

            # Add schema record
            schema_id = "GHjSbphAcdsrZrLjSvsjMp:2:faber-simple:1.1"
            schema_id_parts = schema_id.split(":")
            schema_tags = {
                "schema_id": schema_id,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "epoch": str(int(time())),
            }
            await storage.add_record(
                StorageRecord(SCHEMA_SENT_RECORD_TYPE, schema_id, schema_tags)
            )

            # Add cred def record
            credential_definition_id = "GHjSbphAcdsrZrLjSvsjMp:3:CL:8:default"
            cred_def_tags = {
                "schema_id": schema_id,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "issuer_did": "GHjSbphAcdsrZrLjSvsjMp",
                "cred_def_id": credential_definition_id,
                "epoch": str(int(time())),
            }
            await storage.add_record(
                StorageRecord(
                    CRED_DEF_SENT_RECORD_TYPE, credential_definition_id, cred_def_tags
                )
            )

            # Add private cred def record
            await storage.add_record(
                StorageRecord(
                    CATEGORY_CRED_DEF_PRIVATE,
                    b"{}",
                    {},
                    id=credential_definition_id,
                )
            )

            # Add key proof cred def record
            await storage.add_record(
                StorageRecord(
                    CATEGORY_CRED_DEF_KEY_PROOF,
                    b"{}",
                    {},
                    id=credential_definition_id,
                )
            )

            anoncreds_upgrade.IndyLedgerRequestsExecutor = mock.MagicMock(
                IndyLedgerRequestsExecutor, autospec=True
            )
            anoncreds_upgrade.IndyLedgerRequestsExecutor.return_value.get_ledger_for_identifier = mock.CoroutineMock(
                return_value=(
                    None,
                    mock.MagicMock(
                        get_schema=mock.CoroutineMock(
                            return_value={
                                "attrNames": [
                                    "name",
                                    "age",
                                ],
                            },
                        ),
                        get_credential_definition=mock.CoroutineMock(
                            return_value={
                                "type": "CL",
                                "tag": "default",
                                "value": {
                                    "primary": {
                                        "n": "123",
                                    },
                                },
                            },
                        ),
                    ),
                )
            )

        # Need to update private key with a real private key
        with self.assertRaises(AskarError) as err:
            await anoncreds_upgrade.convert_records_to_anoncreds(self.profile)

        assert "Error updating existing entry" in str(err.exception)

    async def test_retry_converting_records(self):
        with mock.patch.object(
            anoncreds_upgrade, "convert_records_to_anoncreds", mock.CoroutineMock()
        ) as mock_convert_records_to_anoncreds:
            mock_convert_records_to_anoncreds.side_effect = [
                Exception("Error"),
                Exception("Error"),
                None,
            ]
            async with self.profile.session() as session:
                storage = session.inject(BaseStorage)
                upgrading_record = StorageRecord(
                    RECORD_TYPE_ACAPY_UPGRADING,
                    anoncreds_upgrade.UPGRADING_RECORD_IN_PROGRESS,
                )
                await storage.add_record(upgrading_record)

            await anoncreds_upgrade.retry_converting_records(
                self.profile, upgrading_record, 0
            )

            assert mock_convert_records_to_anoncreds.call_count == 3

    async def test_upgrade_wallet_to_anoncreds(self):
        # upgrading record not present
        await anoncreds_upgrade.upgrade_wallet_to_anoncreds_if_requested(self.profile)

    async def test_set_storage_type_to_anoncreds_no_existing_record(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            await storage.add_record(
                StorageRecord(
                    RECORD_TYPE_ACAPY_UPGRADING,
                    "true",
                )
            )
            await anoncreds_upgrade.finish_upgrade(self.profile)

            storage_type_record = await storage.find_record(
                RECORD_TYPE_ACAPY_STORAGE_TYPE, tag_query={}
            )
            assert storage_type_record.value == STORAGE_TYPE_VALUE_ANONCREDS

    async def test_set_storage_type_to_anoncreds_has_existing_record(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            await storage.add_record(
                StorageRecord(
                    RECORD_TYPE_ACAPY_STORAGE_TYPE,
                    "askar",
                )
            )
            await storage.add_record(
                StorageRecord(
                    RECORD_TYPE_ACAPY_UPGRADING,
                    "true",
                )
            )
            await anoncreds_upgrade.finish_upgrade(self.profile)
            storage_type_record = await storage.find_record(
                RECORD_TYPE_ACAPY_STORAGE_TYPE, tag_query={}
            )
            assert storage_type_record.value == STORAGE_TYPE_VALUE_ANONCREDS

    async def test_update_if_subwallet_and_set_storage_type_with_subwallet(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            await storage.add_record(
                StorageRecord(
                    RECORD_TYPE_ACAPY_STORAGE_TYPE,
                    "askar",
                )
            )
            await anoncreds_upgrade.finish_upgrade_by_updating_profile_or_shutting_down(
                self.profile, True
            )
            storage_type_record = await storage.find_record(
                RECORD_TYPE_ACAPY_STORAGE_TYPE, tag_query={}
            )
            assert storage_type_record.value == STORAGE_TYPE_VALUE_ANONCREDS
            assert self.profile.context.injector.get_provider(
                BaseCache
            )._instance.flush.called

    async def test_update_if_subwallet_and_set_storage_type_with_base_wallet(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            await storage.add_record(
                StorageRecord(
                    RECORD_TYPE_ACAPY_STORAGE_TYPE,
                    "askar",
                )
            )
        await anoncreds_upgrade.finish_upgrade_by_updating_profile_or_shutting_down(
            self.profile, False
        )

    async def test_failed_upgrade(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)

            schema_id = "GHjSbphAcdsrZrLjSvsjMp:2:faber-simple:1.1"
            schema_id_parts = schema_id.split(":")
            schema_tags = {
                "schema_id": schema_id,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "epoch": str(int(time())),
            }
            await storage.add_record(
                StorageRecord(SCHEMA_SENT_RECORD_TYPE, schema_id, schema_tags)
            )
            await storage.add_record(
                StorageRecord(
                    RECORD_TYPE_ACAPY_STORAGE_TYPE,
                    "askar",
                )
            )
            await storage.add_record(
                StorageRecord(
                    RECORD_TYPE_ACAPY_UPGRADING,
                    "true",
                )
            )

            credential_definition_id = "GHjSbphAcdsrZrLjSvsjMp:3:CL:8:default"
            cred_def_tags = {
                "schema_id": schema_id,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "issuer_did": "GHjSbphAcdsrZrLjSvsjMp",
                "cred_def_id": credential_definition_id,
                "epoch": str(int(time())),
            }
            await storage.add_record(
                StorageRecord(
                    CRED_DEF_SENT_RECORD_TYPE, credential_definition_id, cred_def_tags
                )
            )
            # Add private cred def record
            await storage.add_record(
                StorageRecord(
                    CATEGORY_CRED_DEF_PRIVATE,
                    b"{}",
                    {},
                    id=credential_definition_id,
                )
            )

            # Add key proof cred def record
            await storage.add_record(
                StorageRecord(
                    CATEGORY_CRED_DEF_KEY_PROOF,
                    b"{}",
                    {},
                    id=credential_definition_id,
                )
            )
            anoncreds_upgrade.IndyLedgerRequestsExecutor = mock.MagicMock()
            anoncreds_upgrade.IndyLedgerRequestsExecutor.return_value.get_ledger_for_identifier = mock.CoroutineMock(
                return_value=(
                    None,
                    mock.MagicMock(
                        get_schema=mock.CoroutineMock(
                            return_value={
                                "attrNames": [
                                    "name",
                                    "age",
                                ],
                            },
                        ),
                        get_credential_definition=mock.CoroutineMock(
                            return_value={
                                "type": "CL",
                                "tag": "default",
                                "value": {
                                    "primary": {
                                        "n": "123",
                                    },
                                },
                            },
                        ),
                    ),
                )
            )

            with (
                mock.patch.object(anoncreds_upgrade, "upgrade_and_delete_schema_records"),
                mock.patch.object(
                    anoncreds_upgrade, "upgrade_and_delete_cred_def_records"
                ),
                mock.patch.object(AskarProfileSession, "rollback") as mock_rollback,
                mock.patch.object(
                    AskarProfileSession,
                    "commit",
                    # Don't wait for sleep in retry to speed up test
                ) as mock_commit,
                mock.patch.object(asyncio, "sleep"),
            ):
                """
                Only tests schemas and cred_defs failing to upgrade because the other objects are
                hard to mock. These tests should be enough to cover them as the logic is the same.
                """

                # Schemas fails to upgrade
                anoncreds_upgrade.upgrade_and_delete_schema_records = mock.CoroutineMock(
                    # Needs to fail 5 times because of the retry logic
                    side_effect=[
                        Exception("Error"),
                        Exception("Error"),
                        Exception("Error"),
                        Exception("Error"),
                        Exception("Error"),
                    ]
                )
                await anoncreds_upgrade.upgrade_wallet_to_anoncreds_if_requested(
                    self.profile
                )
                assert mock_rollback.called
                assert not mock_commit.called
                # Upgrading record should not be deleted
                with self.assertRaises(Exception):
                    await storage.find_record(
                        type_filter=RECORD_TYPE_ACAPY_UPGRADING, tag_query={}
                    )

                storage_type_record = await storage.find_record(
                    type_filter=RECORD_TYPE_ACAPY_STORAGE_TYPE, tag_query={}
                )
                # Storage type should not be updated
                assert storage_type_record.value == "askar"

                # Cred_defs fails to upgrade
                anoncreds_upgrade.upgrade_and_delete_cred_def_records = (
                    mock.CoroutineMock(
                        side_effect=[
                            Exception("Error"),
                            Exception("Error"),
                            Exception("Error"),
                            Exception("Error"),
                            Exception("Error"),
                        ]
                    )
                )
                await anoncreds_upgrade.upgrade_wallet_to_anoncreds_if_requested(
                    self.profile
                )
                assert mock_rollback.called
                assert not mock_commit.called
                # Upgrading record should not be deleted
                with self.assertRaises(Exception):
                    await storage.find_record(
                        type_filter=RECORD_TYPE_ACAPY_UPGRADING, tag_query={}
                    )

                storage_type_record = await storage.find_record(
                    type_filter=RECORD_TYPE_ACAPY_STORAGE_TYPE, tag_query={}
                )
                # Storage type should not be updated
                assert storage_type_record.value == "askar"



from .options import DataIntegrityProofOptions, DataIntegrityProofOptionsSchema
from .proof import DataIntegrityProof, DataIntegrityProofSchema
from .verification_response import (
    DataIntegrityVerificationResponse,
    DataIntegrityVerificationResponseSchema,
)

__all__ = [
    "DataIntegrityProof",
    "DataIntegrityProofSchema",
    "DataIntegrityProofOptions",
    "DataIntegrityProofOptionsSchema",
    "DataIntegrityVerificationResponse",
    "DataIntegrityVerificationResponseSchema",
]



"""Merkle tree hasher for leaf and children nodes."""

import hashlib
from binascii import hexlify, unhexlify


class TreeHasher(object):
    """Merkle tree hasher for bytes data."""

    def __init__(self, hashfunc=hashlib.sha256):
        """Initialize TreeHasher."""
        self.hashfunc = hashfunc

    def hash_leaf(self, data):
        """Return leaf node hash."""
        hasher = self.hashfunc()
        hasher.update(b"\x00" + data)
        return hasher.digest()

    def hash_children(self, left, right):
        """Return parent node hash corresponding to 2 child nodes."""
        hasher = self.hashfunc()
        hasher.update(b"\x01" + left + right)
        return hasher.digest()


class HexTreeHasher(TreeHasher):
    """Merkle tree hasher for hex data."""

    def __init__(self, hashfunc=hashlib.sha256):
        """Initialize HexTreeHasher."""
        self.hasher = TreeHasher(hashfunc)

    def hash_leaf(self, data):
        """Return leaf node hash."""
        return hexlify(self.hasher.hash_leaf(unhexlify(data)))

    def hash_children(self, left, right):
        """Return parent node hash corresponding to 2 child nodes."""
        return hexlify(self.hasher.hash_children(unhexlify(left), unhexlify(right)))



"""present-proof-v2 format handler - supports ANONCREDS, DIF and INDY."""

import logging
from abc import ABC, abstractmethod
from typing import Optional, Tuple

from .....core.error import BaseError
from .....core.profile import Profile
from .....messaging.decorators.attach_decorator import AttachDecorator
from ..messages.pres import V20Pres
from ..messages.pres_format import V20PresFormat
from ..models.pres_exchange import V20PresExRecord

LOGGER = logging.getLogger(__name__)

PresFormatAttachment = Tuple[V20PresFormat, AttachDecorator]


class V20PresFormatHandlerError(BaseError):
    """Presentation exchange format error under present-proof protocol v2.0."""


class V20PresFormatHandler(ABC):
    """Base Presentation Exchange Handler."""

    format: V20PresFormat.Format = None

    def __init__(self, profile: Profile):
        """Initialize PresExchange Handler."""
        super().__init__()
        self._profile = profile

    @property
    def profile(self) -> Profile:
        """Accessor for the current profile instance.

        Returns:
            The profile instance for this presentation exchange format

        """
        return self._profile

    @abstractmethod
    def get_format_identifier(self, message_type: str) -> str:
        """Get attachment format identifier for format and message combination.

        Args:
            message_type (str): Message type for which to return the format identifier

        Returns:
            str: Issue credential attachment format identifier

        """

    @abstractmethod
    def get_format_data(self, message_type: str, data: dict) -> PresFormatAttachment:
        """Get presentation format and attach objects for use in pres_ex messages."""

    @classmethod
    @abstractmethod
    def validate_fields(cls, message_type: str, attachment_data: dict) -> None:
        """Validate attachment data for specific message type and format."""

    @abstractmethod
    async def create_bound_request(
        self,
        pres_ex_record: V20PresExRecord,
        request_data: Optional[dict] = None,
    ) -> PresFormatAttachment:
        """Create a presentation request bound to a proposal."""

    @abstractmethod
    async def create_pres(
        self,
        pres_ex_record: V20PresExRecord,
        request_data: Optional[dict] = None,
    ) -> PresFormatAttachment:
        """Create a presentation."""

    @abstractmethod
    async def receive_pres(self, message: V20Pres, pres_ex_record: V20PresExRecord):
        """Receive a presentation, from message in context on manager creation."""

    @abstractmethod
    async def verify_pres(self, pres_ex_record: V20PresExRecord) -> V20PresExRecord:
        """Verify a presentation."""



"""Handler for incoming invitation request messages."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ....out_of_band.v1_0.manager import OutOfBandManager
from ....out_of_band.v1_0.messages.invitation import HSProto
from ..messages.invitation import Invitation as IntroInvitation
from ..messages.invitation_request import InvitationRequest as IntroInvitationRequest


class InvitationRequestHandler(BaseHandler):
    """Handler for incoming invitation request messages."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler implementation."""
        self._logger.debug("InvitationRequestHandler called with context %s", context)
        assert isinstance(context.message, IntroInvitationRequest)

        if not context.connection_ready:
            raise HandlerException(
                "No connection established for introduction invitation request message"
            )

        # Need a way to prompt the user for acceptance?

        if context.settings.get("auto_accept_intro_invitation_requests"):
            # Create a new connection invitation and send it back in an IntroInvitation
            profile = context.profile
            mgr = OutOfBandManager(profile)
            invite = await mgr.create_invitation(
                use_did_method="did:peer:4",
                hs_protos=[HSProto.DIDEX_1_1],
            )
            response = IntroInvitation(
                invitation=invite.invitation, message=context.message.message
            )
            response.assign_thread_from(context.message)
            response.assign_trace_from(context.message)
            await responder.send_reply(response)



"""Message type identifiers for Feature Discovery."""

from ...didcomm_prefix import DIDCommPrefix

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/tree/"
    "b3a3942ef052039e73cd23d847f42947f8287da2/features/0031-discover-features"
)

# Message types
DISCLOSE = "discover-features/1.0/disclose"
QUERY = "discover-features/1.0/query"

PROTOCOL_PACKAGE = "acapy_agent.protocols.discovery.v1_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        DISCLOSE: f"{PROTOCOL_PACKAGE}.messages.disclose.Disclose",
        QUERY: f"{PROTOCOL_PACKAGE}.messages.query.Query",
    }
)



"""Credential request message handler."""

from .....anoncreds.issuer import AnonCredsIssuerError
from .....core.oob_processor import OobMessageProcessor
from .....indy.issuer import IndyIssuerError
from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError, StorageNotFoundError
from .....utils.tracing import get_timer, trace_event
from .. import problem_report_for_record
from ..manager import V20CredManager, V20CredManagerError
from ..messages.cred_problem_report import ProblemReportReason
from ..messages.cred_request import V20CredRequest


class V20CredRequestHandler(BaseHandler):
    """Message handler class for credential requests."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for credential requests.

        Args:
            context: request context
            responder: responder callback

        """
        r_time = get_timer()

        self._logger.debug("V20CredRequestHandler called with context %s", context)
        assert isinstance(context.message, V20CredRequest)
        self._logger.debug(
            "Received v2.0 credential request message: %s",
            context.message.serialize(as_string=True),
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for credential request not ready")

        # Find associated oob record. If the credential offer was created as an oob
        # attachment the presentation exchange record won't have a connection id (yet)
        oob_processor = context.inject(OobMessageProcessor)
        oob_record = await oob_processor.find_oob_record_for_inbound_message(context)

        # Either connection or oob context must be present
        if not context.connection_record and not oob_record:
            raise HandlerException(
                "No connection or associated connectionless exchange found for credential"
                " request"
            )

        profile = context.profile
        cred_manager = V20CredManager(profile)
        try:
            cred_ex_record = await cred_manager.receive_request(
                context.message, context.connection_record, oob_record
            )  # mgr only finds, saves record: on exception, saving state null is hopeless
        except StorageNotFoundError:
            # issue a problem report...
            cred_ex_record = None
            thread_id = context.message._thread_id
            await responder.send_reply(
                problem_report_for_record(
                    None,
                    ProblemReportReason.RECORD_NOT_FOUND.value,
                    thread_id=thread_id,
                )
            )

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="V20CredRequestHandler.handle.END",
            perf_counter=r_time,
        )

        # If auto_issue is enabled, respond immediately
        if cred_ex_record and cred_ex_record.auto_issue:
            cred_issue_message = None
            try:
                (
                    cred_ex_record,
                    cred_issue_message,
                ) = await cred_manager.issue_credential(
                    cred_ex_record=cred_ex_record,
                    comment=context.message.comment,
                )
                await responder.send_reply(cred_issue_message)
            except (
                BaseModelError,
                AnonCredsIssuerError,
                IndyIssuerError,
                LedgerError,
                StorageError,
                V20CredManagerError,
            ) as err:
                self._logger.exception(err)
                async with profile.session() as session:
                    await cred_ex_record.save_error_state(
                        session,
                        reason=err.roll_up,  # us be specific
                    )
                await responder.send_reply(
                    problem_report_for_record(
                        cred_ex_record,
                        ProblemReportReason.ISSUANCE_ABANDONED.value,  # them: vague
                    )
                )

            trace_event(
                context.settings,
                cred_issue_message,
                outcome="V20CredRequestHandler.issue.END",
                perf_counter=r_time,
            )



"""Test keylist message."""

from unittest import TestCase

from ...message_types import KEYLIST
from ..inner.keylist_key import KeylistKey
from ..inner.keylist_query_paginate import KeylistQueryPaginate
from ..keylist import Keylist, KeylistSchema
from . import MessageTest


class TestKeylist(MessageTest, TestCase):
    """Test Keylist message."""

    TYPE = KEYLIST
    CLASS = Keylist
    SCHEMA = KeylistSchema
    VALUES = {
        "pagination": KeylistQueryPaginate(10, 10),
        "keys": [
            KeylistKey(
                recipient_key="did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                action="added",
                result="success",
            )
        ],
    }



"""Classes to manage presentations."""

import logging
from typing import Optional, Tuple

from ....connections.models.conn_record import ConnRecord
from ....core.error import BaseError
from ....core.profile import Profile
from ....messaging.responder import BaseResponder
from ...out_of_band.v1_0.models.oob_record import OobRecord
from .messages.pres import V20Pres
from .messages.pres_ack import V20PresAck
from .messages.pres_format import V20PresFormat
from .messages.pres_problem_report import ProblemReportReason, V20PresProblemReport
from .messages.pres_proposal import V20PresProposal
from .messages.pres_request import V20PresRequest
from .models.pres_exchange import V20PresExRecord

LOGGER = logging.getLogger(__name__)


class V20PresManagerError(BaseError):
    """Presentation error."""


class V20PresManager:
    """Class for managing presentations."""

    def __init__(self, profile: Profile):
        """Initialize a V20PresManager.

        Args:
            profile: The profile instance for this presentation manager
        """

        self._profile = profile

    async def create_exchange_for_proposal(
        self,
        connection_id: str,
        pres_proposal_message: V20PresProposal,
        auto_present: Optional[bool] = None,
        auto_remove: Optional[bool] = None,
    ):
        """Create a presentation exchange record for input presentation proposal.

        Args:
            connection_id: connection identifier
            pres_proposal_message: presentation proposal to serialize
                to exchange record
            auto_present: whether to present proof upon receiving proof request
                (default to configuration setting)
            auto_remove: whether to remove this presentation exchange upon completion

        Returns:
            Presentation exchange record, created

        """
        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        pres_ex_record = V20PresExRecord(
            connection_id=connection_id,
            thread_id=pres_proposal_message._thread_id,
            initiator=V20PresExRecord.INITIATOR_SELF,
            role=V20PresExRecord.ROLE_PROVER,
            state=V20PresExRecord.STATE_PROPOSAL_SENT,
            pres_proposal=pres_proposal_message,
            auto_present=auto_present,
            trace=(pres_proposal_message._trace is not None),
            auto_remove=auto_remove,
        )

        async with self._profile.session() as session:
            await pres_ex_record.save(session, reason="create v2.0 presentation proposal")

        return pres_ex_record

    async def receive_pres_proposal(
        self, message: V20PresProposal, conn_record: ConnRecord
    ):
        """Receive a presentation proposal from message in context on manager creation.

        Returns:
            Presentation exchange record, created

        """
        pres_ex_record = V20PresExRecord(
            connection_id=conn_record.connection_id,
            thread_id=message._thread_id,
            initiator=V20PresExRecord.INITIATOR_EXTERNAL,
            role=V20PresExRecord.ROLE_VERIFIER,
            state=V20PresExRecord.STATE_PROPOSAL_RECEIVED,
            pres_proposal=message,
            trace=(message._trace is not None),
            auto_remove=not self._profile.settings.get("preserve_exchange_records"),
        )

        async with self._profile.session() as session:
            await pres_ex_record.save(session, reason="receive v2.0 presentation request")

        return pres_ex_record

    async def create_bound_request(
        self,
        pres_ex_record: V20PresExRecord,
        request_data: Optional[dict] = None,
        comment: Optional[str] = None,
    ):
        """Create a presentation request bound to a proposal.

        Args:
            pres_ex_record: Presentation exchange record for which
                to create presentation request
            request_data: Formatted requested attributes and predicates
            comment: Optional human-readable comment pertaining to request creation

        Returns:
            A tuple (updated presentation exchange record, presentation request message)
        """
        proof_proposal = pres_ex_record.pres_proposal
        input_formats = proof_proposal.formats
        request_formats = []
        for format in input_formats:
            pres_exch_format = V20PresFormat.Format.get(format.format)

            if pres_exch_format:
                request_formats.append(
                    await pres_exch_format.handler(self._profile).create_bound_request(
                        pres_ex_record,
                        request_data,
                    )
                )
        if len(request_formats) == 0:
            raise V20PresManagerError(
                "Unable to create presentation request. No supported formats"
            )
        pres_request_message = V20PresRequest(
            comment=comment,
            will_confirm=True,
            formats=[format for (format, _) in request_formats],
            request_presentations_attach=[attach for (_, attach) in request_formats],
        )
        pres_request_message._thread = {"thid": pres_ex_record.thread_id}
        pres_request_message.assign_trace_decorator(
            self._profile.settings, pres_ex_record.trace
        )

        pres_ex_record.thread_id = pres_request_message._thread_id
        pres_ex_record.state = V20PresExRecord.STATE_REQUEST_SENT
        pres_ex_record.pres_request = pres_request_message
        async with self._profile.session() as session:
            await pres_ex_record.save(
                session, reason="create (bound) v2.0 presentation request"
            )

        return pres_ex_record, pres_request_message

    async def create_exchange_for_request(
        self,
        connection_id: str,
        pres_request_message: V20PresRequest,
        auto_verify: Optional[bool] = None,
        auto_remove: Optional[bool] = None,
    ):
        """Create a presentation exchange record for input presentation request.

        Args:
            connection_id: connection identifier
            pres_request_message: presentation request to use in creating
                exchange record, extracting indy proof request and thread id
            auto_verify: whether to auto-verify presentation exchange
            auto_remove: whether to remove this presentation exchange upon completion

        Returns:
            Presentation exchange record, updated

        """
        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        pres_ex_record = V20PresExRecord(
            connection_id=connection_id,
            thread_id=pres_request_message._thread_id,
            initiator=V20PresExRecord.INITIATOR_SELF,
            role=V20PresExRecord.ROLE_VERIFIER,
            state=V20PresExRecord.STATE_REQUEST_SENT,
            pres_request=pres_request_message,
            auto_verify=auto_verify,
            trace=(pres_request_message._trace is not None),
            auto_remove=auto_remove,
        )
        async with self._profile.session() as session:
            await pres_ex_record.save(
                session, reason="create (free) v2.0 presentation request"
            )

        return pres_ex_record

    async def receive_pres_request(self, pres_ex_record: V20PresExRecord):
        """Receive a presentation request.

        Args:
            pres_ex_record: presentation exchange record with request to receive

        Returns:
            The presentation exchange record, updated

        """
        pres_ex_record.state = V20PresExRecord.STATE_REQUEST_RECEIVED
        async with self._profile.session() as session:
            await pres_ex_record.save(session, reason="receive v2.0 presentation request")

        return pres_ex_record

    async def create_pres(
        self,
        pres_ex_record: V20PresExRecord,
        request_data: Optional[dict] = None,
        *,
        comment: Optional[str] = None,
    ) -> Tuple[V20PresExRecord, V20Pres]:
        """Create a presentation.

        Args:
            pres_ex_record (V20PresExRecord): The record to update.
            request_data (Optional[dict]): Optional indy formatted
                requested_credentials.
            comment (str): Optional human-readable comment.

        Example `request_data` format, mapping proof request referents (uuid)
        to wallet referents (cred id):

        ::

            {
                "self_attested_attributes": {
                    "j233ffbc-bd35-49b1-934f-51e083106f6d": "value"
                },
                "requested_attributes": {
                    "6253ffbb-bd35-49b3-934f-46e083106f6c": {
                        "cred_id": "5bfa40b7-062b-4ae0-a251-a86c87922c0e",
                        "revealed": true
                    }
                },
                "requested_predicates": {
                    "bfc8a97d-60d3-4f21-b998-85eeabe5c8c0": {
                        "cred_id": "5bfa40b7-062b-4ae0-a251-a86c87922c0e"
                    }
                }
            }

        Returns:
            Tuple[V20PresExRecord, V20Pres]: A tuple containing the updated
                presentation exchange record and the presentation message.

        Raises:
            V20PresManagerError: If unable to create the presentation or no supported
                formats are available.
        """

        proof_request = pres_ex_record.pres_request
        input_formats = proof_request.formats
        request_data = request_data or {}
        pres_formats = []
        for format in input_formats:
            pres_exch_format = V20PresFormat.Format.get(format.format)

            if pres_exch_format:
                if not request_data:
                    request_data_pres_exch = {}
                else:
                    request_data_pres_exch = {
                        pres_exch_format.api: request_data.get(pres_exch_format.api)
                    }
                pres_tuple = await pres_exch_format.handler(self._profile).create_pres(
                    pres_ex_record,
                    request_data_pres_exch,
                )
                if pres_tuple:
                    pres_formats.append(pres_tuple)
                else:
                    raise V20PresManagerError(
                        "Unable to create presentation. ProblemReport message sent"
                    )
        if len(pres_formats) == 0:
            raise V20PresManagerError(
                "Unable to create presentation. No supported formats"
            )
        pres_message = V20Pres(
            comment=comment,
            formats=[format for (format, _) in pres_formats],
            presentations_attach=[attach for (_, attach) in pres_formats],
        )

        # Assign thid (and optionally pthid) to message
        pres_message.assign_thread_from(pres_ex_record.pres_request)
        pres_message.assign_trace_decorator(self._profile.settings, pres_ex_record.trace)

        # save presentation exchange state
        pres_ex_record.state = V20PresExRecord.STATE_PRESENTATION_SENT
        pres_ex_record.pres = V20Pres(
            formats=[format for (format, _) in pres_formats],
            presentations_attach=[attach for (_, attach) in pres_formats],
        )
        async with self._profile.session() as session:
            await pres_ex_record.save(session, reason="create v2.0 presentation")
        return pres_ex_record, pres_message

    async def receive_pres(
        self,
        message: V20Pres,
        connection_record: Optional[ConnRecord],
        oob_record: Optional[OobRecord],
    ):
        """Receive a presentation, from message in context on manager creation.

        Returns:
            presentation exchange record, retrieved and updated

        """

        thread_id = message._thread_id
        # Normally we only set the connection_id to None if an oob record is present
        # But present proof supports the old-style AIP-1 connectionless exchange that
        # bypasses the oob record. So we can't verify if an oob record is associated with
        # the exchange because it is possible that there is None
        connection_id = (
            None
            if oob_record
            else connection_record.connection_id
            if connection_record
            else None
        )

        async with self._profile.session() as session:
            pres_ex_record = await V20PresExRecord.retrieve_by_tag_filter(
                session,
                {"thread_id": thread_id},
                {
                    "role": V20PresExRecord.ROLE_VERIFIER,
                    "connection_id": connection_id,
                },
            )

        # Save connection id (if it wasn't already present)
        if connection_record:
            pres_ex_record.connection_id = connection_record.connection_id

        input_formats = message.formats

        for format in input_formats:
            pres_format = V20PresFormat.Format.get(format.format)

            if pres_format:
                receive_pres_return = await pres_format.handler(
                    self._profile
                ).receive_pres(
                    message,
                    pres_ex_record,
                )
                if isinstance(receive_pres_return, bool) and not receive_pres_return:
                    raise V20PresManagerError(
                        "Unable to verify received presentation."
                        " ProblemReport message sent"
                    )
        pres_ex_record.pres = message
        pres_ex_record.state = V20PresExRecord.STATE_PRESENTATION_RECEIVED
        async with self._profile.session() as session:
            await pres_ex_record.save(session, reason="receive v2.0 presentation")

        return pres_ex_record

    async def verify_pres(
        self, pres_ex_record: V20PresExRecord, responder: Optional[BaseResponder] = None
    ):
        """Verify a presentation.

        Args:
            pres_ex_record: presentation exchange record
                with presentation request and presentation to verify
            responder: base responder

        Returns:
            presentation exchange record, updated

        """
        pres_request_msg = pres_ex_record.pres_request
        input_formats = pres_request_msg.formats
        for format in input_formats:
            pres_exch_format = V20PresFormat.Format.get(format.format)

            if pres_exch_format:
                pres_ex_record = await pres_exch_format.handler(
                    self._profile
                ).verify_pres(
                    pres_ex_record,
                )
                if pres_ex_record.verified == "false":
                    break

        pres_ex_record.state = V20PresExRecord.STATE_DONE

        async with self._profile.session() as session:
            await pres_ex_record.save(session, reason="verify v2.0 presentation")

        if pres_request_msg.will_confirm:
            await self.send_pres_ack(pres_ex_record, responder)

        return pres_ex_record

    async def send_pres_ack(
        self, pres_ex_record: V20PresExRecord, responder: Optional[BaseResponder] = None
    ):
        """Send acknowledgement of presentation receipt.

        Args:
            pres_ex_record: presentation exchange record with thread id
            responder: base responder

        """
        responder = responder or self._profile.inject_or(BaseResponder)

        if responder:
            pres_ack_message = V20PresAck(verification_result=pres_ex_record.verified)
            pres_ack_message._thread = {"thid": pres_ex_record.thread_id}
            pres_ack_message.assign_trace_decorator(
                self._profile.settings, pres_ex_record.trace
            )

            await responder.send_reply(
                pres_ack_message,
                # connection_id can be none in case of connectionless
                connection_id=pres_ex_record.connection_id,
            )

            # all done: delete
            if pres_ex_record.auto_remove:
                async with self._profile.session() as session:
                    await pres_ex_record.delete_record(session)
        else:
            LOGGER.warning(
                "Configuration has no BaseResponder: cannot ack presentation on %s",
                pres_ex_record.thread_id,
            )

    async def receive_pres_ack(self, message: V20PresAck, conn_record: ConnRecord):
        """Receive a presentation ack, from message in context on manager creation.

        Returns:
            presentation exchange record, retrieved and updated

        """
        connection_id = conn_record.connection_id if conn_record else None
        async with self._profile.session() as session:
            pres_ex_record = await V20PresExRecord.retrieve_by_tag_filter(
                session,
                {"thread_id": message._thread_id},
                {
                    # connection_id can be null in connectionless
                    "connection_id": connection_id,
                    "role": V20PresExRecord.ROLE_PROVER,
                },
            )
            pres_ex_record.verified = message._verification_result
            pres_ex_record.state = V20PresExRecord.STATE_DONE

            await pres_ex_record.save(session, reason="receive v2.0 presentation ack")

            # all done: delete
            if pres_ex_record.auto_remove:
                async with self._profile.session() as session:
                    await pres_ex_record.delete_record(session)

        return pres_ex_record

    async def receive_problem_report(
        self, message: V20PresProblemReport, connection_id: str
    ):
        """Receive problem report.

        Returns:
            presentation exchange record, retrieved and updated

        """
        # FIXME use transaction, fetch for_update
        async with self._profile.session() as session:
            pres_ex_record = await V20PresExRecord.retrieve_by_tag_filter(
                session,
                {"thread_id": message._thread_id},
                {"connection_id": connection_id},
            )

            pres_ex_record.state = V20PresExRecord.STATE_ABANDONED
            code = message.description.get("code", ProblemReportReason.ABANDONED.value)
            pres_ex_record.error_msg = f"{code}: {message.description.get('en', code)}"
            await pres_ex_record.save(session, reason="received problem report")

        return pres_ex_record



import json
import tempfile
from unittest import IsolatedAsyncioTestCase

import pytest

from ....ledger.base import BaseLedger
from ....ledger.multiple_ledger.ledger_requests_executor import IndyLedgerRequestsExecutor
from ....tests import mock
from ....utils.testing import create_test_profile
from .. import holder, issuer, verifier

TEST_DID = "55GkHamhTU1ZbTbV2ab9DE"
SCHEMA_NAME = "resident"
SCHEMA_VERSION = "1.0"
SCHEMA_TXN = 1234
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:{SCHEMA_VERSION}"
CRED_DEF_ID = f"{TEST_DID}:3:CL:{SCHEMA_TXN}:default"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:0"
CRED_REFT = "attr_0_uuid"
PRES_REQ_NON_REV = {
    "name": "pres-request",
    "version": "1.0",
    "nonce": "1234567890",
    "requested_attributes": {CRED_REFT: {"names": ["name", "moniker"]}},
    "requested_predicates": {},
}
TIMESTAMP = 99999
PRES_REQ_REV = {
    "name": "pres-request",
    "version": "1.0",
    "nonce": "1234567890",
    "requested_attributes": {CRED_REFT: {"names": ["name", "moniker"]}},
    "requested_predicates": {},
    "non_revoked": {"to": TIMESTAMP},
}


@pytest.mark.askar
@pytest.mark.indy_credx
class TestIndyCredxIssuance(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"value": {}}
        )
        mock_ledger.get_revoc_reg_delta = mock.CoroutineMock(
            return_value=(
                {"value": {"...": "..."}},
                1234567890,
            )
        )

        self.ledger = mock_ledger

        self.holder_profile = await create_test_profile()
        self.issuer_profile = await create_test_profile()
        self.issuer_profile._context.injector.bind_instance(BaseLedger, self.ledger)

        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.issuer_profile._context.injector.bind_instance(
            IndyLedgerRequestsExecutor, mock_executor
        )

        self.holder = holder.IndyCredxHolder(self.holder_profile)
        self.issuer = issuer.IndyCredxIssuer(self.issuer_profile)
        self.verifier = verifier.IndyCredxVerifier(self.issuer_profile)
        assert "IndyCredxHolder" in str(self.holder)
        assert "IndyCredxIssuer" in str(self.issuer)
        assert "IndyCredxVerifier" in str(self.verifier)

    async def test_issue_store_non_rev(self):
        assert (
            self.issuer.make_schema_id(TEST_DID, SCHEMA_NAME, SCHEMA_VERSION) == SCHEMA_ID
        )

        (s_id, schema_json) = await self.issuer.create_schema(
            TEST_DID,
            SCHEMA_NAME,
            SCHEMA_VERSION,
            ["name", "moniker"],
        )
        assert s_id == SCHEMA_ID
        schema = json.loads(schema_json)
        schema["seqNo"] = SCHEMA_TXN

        assert (
            self.issuer.make_credential_definition_id(TEST_DID, schema, tag="default")
            == CRED_DEF_ID
        )

        (
            cd_id,
            cred_def_json,
        ) = await self.issuer.create_and_store_credential_definition(
            TEST_DID, schema, support_revocation=False
        )
        assert cd_id == CRED_DEF_ID
        assert await self.issuer.credential_definition_in_wallet(cd_id)
        cred_def = json.loads(cred_def_json)

        cred_offer_json = await self.issuer.create_credential_offer(cd_id)
        cred_offer = json.loads(cred_offer_json)

        cred_req_json, cred_req_meta_json = await self.holder.create_credential_request(
            cred_offer, cred_def, TEST_DID
        )
        cred_req = json.loads(cred_req_json)
        cred_req_meta = json.loads(cred_req_meta_json)

        cred_json, cred_rev_id = await self.issuer.create_credential(
            schema,
            cred_offer,
            cred_req,
            {"name": "NAME", "moniker": "MONIKER"},
            revoc_reg_id=None,
            tails_file_path=None,
        )
        assert cred_rev_id is None
        cred_data = json.loads(cred_json)

        cred_id = await self.holder.store_credential(cred_def, cred_data, cred_req_meta)

        found = await self.holder.get_credential(cred_id)
        assert found
        stored_cred = json.loads(found)

        assert not await self.holder.get_mime_type(cred_id, "name")

        creds = await self.holder.get_credentials(offset=None, limit=None, wql=None)
        assert len(creds) == 1
        assert creds[0] == stored_cred

        assert not await self.holder.credential_revoked(self.ledger, cred_id)

        pres_creds = (
            await self.holder.get_credentials_for_presentation_request_by_referent(
                PRES_REQ_NON_REV,
                None,
                offset=0,
                limit=10,
                extra_query={},
            )
        )
        assert pres_creds == [
            {
                "cred_info": stored_cred,
                "interval": None,
                "presentation_referents": [CRED_REFT],
            }
        ]

        pres_json = await self.holder.create_presentation(
            PRES_REQ_NON_REV,
            {"requested_attributes": {CRED_REFT: {"cred_id": cred_id, "revealed": True}}},
            {s_id: schema},
            {cd_id: cred_def},
            rev_states=None,
        )
        pres = json.loads(pres_json)

        assert await self.verifier.verify_presentation(
            PRES_REQ_NON_REV, pres, {s_id: schema}, {cd_id: cred_def}, {}, {}
        )

        await self.holder.delete_credential(cred_id)

    async def test_issue_store_rev(self):
        assert (
            self.issuer.make_schema_id(TEST_DID, SCHEMA_NAME, SCHEMA_VERSION) == SCHEMA_ID
        )

        (s_id, schema_json) = await self.issuer.create_schema(
            TEST_DID,
            SCHEMA_NAME,
            SCHEMA_VERSION,
            ["name", "moniker"],
        )
        assert s_id == SCHEMA_ID
        schema = json.loads(schema_json)
        schema["seqNo"] = SCHEMA_TXN

        assert (
            self.issuer.make_credential_definition_id(TEST_DID, schema, tag="default")
            == CRED_DEF_ID
        )

        (
            cd_id,
            cred_def_json,
        ) = await self.issuer.create_and_store_credential_definition(
            TEST_DID, schema, support_revocation=True
        )
        assert cd_id == CRED_DEF_ID
        cred_def = json.loads(cred_def_json)
        self.ledger.get_credential_definition.return_value = cred_def

        with tempfile.TemporaryDirectory() as tmp_path:
            (
                reg_id,
                reg_def_json,
                reg_entry_json,
            ) = await self.issuer.create_and_store_revocation_registry(
                TEST_DID, cd_id, "CL_ACCUM", "0", 10, tmp_path
            )
            assert reg_id == REV_REG_ID
            reg_def = json.loads(reg_def_json)
            reg_entry = json.loads(reg_entry_json)
            tails_path = reg_def["value"]["tailsLocation"]

            cred_offer_json = await self.issuer.create_credential_offer(cd_id)
            cred_offer = json.loads(cred_offer_json)

            (
                cred_req_json,
                cred_req_meta_json,
            ) = await self.holder.create_credential_request(
                cred_offer, cred_def, TEST_DID
            )
            cred_req = json.loads(cred_req_json)
            cred_req_meta = json.loads(cred_req_meta_json)

            cred_json, cred_rev_id = await self.issuer.create_credential(
                schema,
                cred_offer,
                cred_req,
                {"name": "NAME", "moniker": "MONIKER"},
                revoc_reg_id=reg_id,
                tails_file_path=tails_path,
            )
            assert cred_rev_id == "1"
            cred_data = json.loads(cred_json)

            cred_id = await self.holder.store_credential(
                cred_def,
                cred_data,
                cred_req_meta,
                rev_reg_def=reg_def,
            )

            found = await self.holder.get_credential(cred_id)
            assert found
            stored_cred = json.loads(found)

            creds = await self.holder.get_credentials(offset=None, limit=None, wql=None)
            assert len(creds) == 1
            assert creds[0] == stored_cred

            assert not await self.holder.credential_revoked(self.ledger, cred_id)

            pres_creds = (
                await self.holder.get_credentials_for_presentation_request_by_referent(
                    PRES_REQ_REV,
                    None,
                    offset=0,
                    limit=10,
                    extra_query={},
                )
            )
            assert pres_creds == [
                {
                    "cred_info": stored_cred,
                    "interval": {"to": TIMESTAMP},
                    "presentation_referents": [CRED_REFT],
                }
            ]

            rev_state_time = 1
            rev_state_json = await self.holder.create_revocation_state(
                cred_rev_id, reg_def, reg_entry, rev_state_time, tails_path
            )
            rev_state_init = json.loads(rev_state_json)
            rev_delta_init = {"ver": "1.0", "value": rev_state_init["rev_reg"]}

            (rev_delta_2_json, skipped_ids) = await self.issuer.revoke_credentials(
                cd_id, reg_id, tails_path, (1,)
            )
            assert not skipped_ids
            rev_delta_2 = json.loads(rev_delta_2_json)

            await self.issuer.merge_revocation_registry_deltas(
                rev_delta_init, rev_delta_2
            )

        pres_json = await self.holder.create_presentation(
            PRES_REQ_REV,
            {
                "requested_attributes": {
                    CRED_REFT: {
                        "cred_id": cred_id,
                        "revealed": True,
                        "timestamp": rev_state_time,
                    }
                }
            },
            {s_id: schema},
            {cd_id: cred_def},
            rev_states={reg_id: {rev_state_time: rev_state_init}},
        )
        pres = json.loads(pres_json)

        reg_def["txnTime"] = rev_state_time
        assert await self.verifier.verify_presentation(
            PRES_REQ_REV,
            pres,
            {s_id: schema},
            {cd_id: cred_def},
            {reg_id: reg_def},
            {reg_id: {rev_state_time: rev_delta_init}},
        )

        await self.holder.delete_credential(cred_id)



"""Linked Data Signature class."""

from abc import ABCMeta, abstractmethod
from datetime import datetime, timezone
from hashlib import sha256
from typing import Optional, Union

from dateutil import tz

from ..constants import SECURITY_CONTEXT_URL
from ..document_loader import DocumentLoaderMethod
from ..error import LinkedDataProofException
from ..purposes import _ProofPurpose as ProofPurpose
from ..validation_result import ProofResult
from .linked_data_proof import LinkedDataProof


class LinkedDataSignature(LinkedDataProof, metaclass=ABCMeta):
    """Linked Data Signature class."""

    def __init__(
        self,
        *,
        proof: Optional[dict] = None,
        verification_method: Optional[str] = None,
        date: Union[datetime, None] = None,
    ):
        """Create new LinkedDataSignature instance.

        Must be subclassed, not initialized directly.

        Args:
            signature_type (str): Signature type to use for the proof
            proof (dict, optional): A JSON-LD document with options to use for the
                `proof` node (e.g. any other custom fields can be provided here
                using a context different from security-v2).
            verification_method (str, optional): A key id URL to the paired public key.
            date (datetime, optional): Signing date to use. Defaults to now

        """
        super().__init__(proof=proof)
        self.verification_method = verification_method
        self.date = date

    @abstractmethod
    async def sign(self, *, verify_data: bytes, proof: dict) -> dict:
        """Sign the data and add it to the proof.

        Args:
            verify_data (bytes): The data to sign.
            proof (dict): The proof to add the signature to

        Returns:
            dict: The proof object with the added signature

        """
        pass

    @abstractmethod
    async def verify_signature(
        self,
        *,
        verify_data: bytes,
        verification_method: dict,
        document: dict,
        proof: dict,
        document_loader: DocumentLoaderMethod,
    ) -> bool:
        """Verify the data against the proof.

        Args:
            verify_data (bytes): The data to check
            verification_method (dict): The verification method to use.
            document (dict): The document the verify data is derived for as extra context
            proof (dict): The proof to check
            document_loader (DocumentLoader): Document loader used for resolving

        Returns:
            bool: Whether the signature is valid for the data

        """

    async def create_proof(
        self,
        *,
        document: dict,
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> dict:
        """Create proof for document, return proof."""
        proof = self.proof.copy() if self.proof else {}

        # TODO: validate if verification_method is set?
        proof["type"] = self.signature_type
        proof["verificationMethod"] = self.verification_method

        # Set created if not already set
        if not proof.get("created"):
            # Use class date, or now
            date = self.date or datetime.now(timezone.utc)
            if not date.tzinfo:
                date = date.replace(tzinfo=tz.UTC)
            proof["created"] = date.isoformat(timespec="seconds")

        # Allow purpose to update the proof; the `proof` is in the
        # SECURITY_CONTEXT_URL `@context` -- therefore the `purpose` must
        # ensure any added fields are also represented in that same `@context`
        proof = purpose.update(proof)

        # Create data to sign
        verify_data = self._create_verify_data(
            proof=proof, document=document, document_loader=document_loader
        )

        # Sign data
        proof = await self.sign(verify_data=verify_data, proof=proof)

        return proof

    async def verify_proof(
        self,
        *,
        proof: dict,
        document: dict,
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> ProofResult:
        """Verify proof against document and proof purpose."""
        try:
            # Create data to verify
            verify_data = self._create_verify_data(
                proof=proof, document=document, document_loader=document_loader
            )

            # Fetch verification method
            verification_method = self._get_verification_method(
                proof=proof, document_loader=document_loader
            )

            # Verify signature on data
            verified = await self.verify_signature(
                verify_data=verify_data,
                verification_method=verification_method,
                document=document,
                proof=proof,
                document_loader=document_loader,
            )
            if not verified:
                raise LinkedDataProofException(
                    f"Invalid signature on document {document}"
                )

            # Ensure proof was performed for a valid purpose
            purpose_result = purpose.validate(
                proof=proof,
                document=document,
                suite=self,
                verification_method=verification_method,
                document_loader=document_loader,
            )

            if not purpose_result.valid:
                return ProofResult(
                    verified=False,
                    purpose_result=purpose_result,
                    error=purpose_result.error,
                )

            return ProofResult(verified=True, purpose_result=purpose_result)
        except Exception as err:
            return ProofResult(verified=False, error=err)

    def _create_verify_data(
        self, *, proof: dict, document: dict, document_loader: DocumentLoaderMethod
    ) -> bytes:
        """Create signing or verification data."""
        c14n_proof_options = self._canonize_proof(
            proof=proof, document=document, document_loader=document_loader
        )
        c14n_doc = self._canonize(input=document, document_loader=document_loader)

        # TODO: detect any dropped properties using expand/contract step

        return (
            sha256(c14n_proof_options.encode("utf-8")).digest()
            + sha256(c14n_doc.encode("utf-8")).digest()
        )

    def _canonize_proof(
        self, *, proof: dict, document: dict, document_loader: DocumentLoaderMethod
    ):
        """Canonize proof dictionary. Removes jws, signature, etc..."""
        # Use default security context url if document has no context
        proof = {
            **proof,
            "@context": document.get("@context") or SECURITY_CONTEXT_URL,
        }

        proof.pop("jws", None)
        proof.pop("signatureValue", None)
        proof.pop("proofValue", None)

        return self._canonize(input=proof, document_loader=document_loader)



"""Normalization methods used while transitioning to DID:Key method."""

from typing import Union

from ....did.did_key import DIDKey
from ....wallet.key_type import ED25519


def normalize_from_did_key(key: str):
    """Normalize Recipient/Routing keys from DID:Key to public keys."""
    if key.startswith("did:key:"):
        return DIDKey.from_did(key).public_key_b58

    return key


def normalize_from_public_key(key: str):
    """Normalize Recipient/Routing keys from public keys to DID:Key."""
    if key.startswith("did:key:"):
        return key

    return DIDKey.from_public_key_b58(key, ED25519).did


def normalize_to_did_key(value: Union[str, DIDKey]) -> DIDKey:
    """Normalize a value to a DIDKey."""
    if isinstance(value, DIDKey):
        return value
    if value.startswith("did:key:"):
        return DIDKey.from_did(value)
    return DIDKey.from_public_key_b58(value, ED25519)



from unittest import IsolatedAsyncioTestCase

from ......core.oob_processor import OobMessageProcessor
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...messages.credential_offer import CredentialOffer
from .. import credential_offer_handler as test_module


class TestCredentialOfferHandler(IsolatedAsyncioTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.settings["debug.auto_respond_credential_offer"] = False
        request_context.connection_record = mock.MagicMock()

        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=mock.MagicMock()
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with mock.patch.object(
            test_module, "CredentialManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_offer = mock.CoroutineMock()
            request_context.message = CredentialOffer()
            request_context.connection_ready = True
            handler = test_module.CredentialOfferHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_cred_mgr.assert_called_once_with(request_context.profile)
        mock_cred_mgr.return_value.receive_offer.assert_called_once_with(
            request_context.message, request_context.connection_record.connection_id
        )
        mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            request_context
        )
        assert not responder.messages

    async def test_called_auto_request(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.settings["debug.auto_respond_credential_offer"] = True
        request_context.connection_record = mock.MagicMock()
        request_context.connection_record.my_did = "dummy"

        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=mock.MagicMock()
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with mock.patch.object(
            test_module, "CredentialManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_offer = mock.CoroutineMock()
            mock_cred_mgr.return_value.create_request = mock.CoroutineMock(
                return_value=(None, "credential_request_message")
            )
            request_context.message = CredentialOffer()
            request_context.connection_ready = True
            handler = test_module.CredentialOfferHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_cred_mgr.assert_called_once_with(request_context.profile)
        mock_cred_mgr.return_value.receive_offer.assert_called_once_with(
            request_context.message, request_context.connection_record.connection_id
        )
        mock_oob_processor.find_oob_record_for_inbound_message.assert_called_once_with(
            request_context
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "credential_request_message"
        assert target == {}

    async def test_called_auto_request_x(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.settings["debug.auto_respond_credential_offer"] = True
        request_context.connection_record = mock.MagicMock()
        request_context.connection_record.my_did = "dummy"

        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=mock.MagicMock()
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with mock.patch.object(
            test_module, "CredentialManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_offer = mock.CoroutineMock(
                return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
            )
            mock_cred_mgr.return_value.create_request = mock.CoroutineMock(
                side_effect=test_module.IndyHolderError()
            )

            request_context.message = CredentialOffer()
            request_context.connection_ready = True
            handler = test_module.CredentialOfferHandler()
            responder = MockResponder()

            with (
                mock.patch.object(responder, "send_reply", mock.CoroutineMock()),
                mock.patch.object(
                    handler._logger, "exception", mock.MagicMock()
                ) as mock_log_exc,
            ):
                await handler.handle(request_context, responder)
                mock_log_exc.assert_called_once()

    async def test_called_not_ready(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "CredentialManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_offer = mock.CoroutineMock()
            request_context.message = CredentialOffer()
            request_context.connection_ready = False
            handler = test_module.CredentialOfferHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException) as err:
                await handler.handle(request_context, responder)
            assert (
                err.exception.message == "Connection used for credential offer not ready"
            )

        assert not responder.messages

    async def test_no_conn_no_oob(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()

        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=None
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with mock.patch.object(
            test_module, "CredentialManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_offer = mock.CoroutineMock()
            request_context.message = CredentialOffer()
            request_context.connection_ready = False
            handler = test_module.CredentialOfferHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException) as err:
                await handler.handle(request_context, responder)
            assert (
                err.exception.message
                == "No connection or associated connectionless exchange found for credential offer"
            )

        assert not responder.messages



"""mediate-deny message used to notify mediation client of a denied mediation request."""

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import MEDIATE_DENY, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.mediation_deny_handler.MediationDenyHandler"


class MediationDeny(AgentMessage):
    """Class representing a mediation deny message."""

    class Meta:
        """Metadata for a mediation deny."""

        handler_class = HANDLER_CLASS
        message_type = MEDIATE_DENY
        schema_class = "MediationDenySchema"

    def __init__(
        self,
        **kwargs,
    ):
        """Initialize mediation deny object."""
        super(MediationDeny, self).__init__(**kwargs)


class MediationDenySchema(AgentMessageSchema):
    """Mediation grant schema class."""

    class Meta:
        """Mediation deny schema metadata."""

        model_class = MediationDeny



"""Ledger base class."""

import json
import logging
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum
from hashlib import sha256
from typing import List, Optional, Sequence, Tuple, Union

from ..indy.issuer import DEFAULT_CRED_DEF_TAG, IndyIssuer, IndyIssuerError
from ..messaging.valid import IndyDID
from ..utils import sentinel
from ..utils.general import strip_did_prefix
from ..wallet.did_info import DIDInfo
from .endpoint_type import EndpointType
from .error import (
    BadLedgerRequestError,
    LedgerError,
    LedgerObjectAlreadyExistsError,
    LedgerTransactionError,
)

LOGGER = logging.getLogger(__name__)


class BaseLedger(ABC, metaclass=ABCMeta):
    """Base class for ledger."""

    BACKEND_NAME: Optional[str] = None

    async def __aenter__(self) -> "BaseLedger":
        """Context manager entry.

        Returns:
            The current instance

        """
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """Context manager exit."""

    @property
    def backend(self) -> str:
        """Accessor for the ledger backend name."""
        return self.__class__.BACKEND_NAME

    @property
    @abstractmethod
    def read_only(self) -> bool:
        """Accessor for the ledger read-only flag."""

    @abstractmethod
    async def is_ledger_read_only(self) -> bool:
        """Check if ledger is read-only including TAA."""

    @abstractmethod
    async def get_key_for_did(self, did: str) -> str:
        """Fetch the verkey for a ledger DID.

        Args:
            did: The DID to look up on the ledger or in the cache
        """

    @abstractmethod
    async def get_endpoint_for_did(
        self, did: str, endpoint_type: EndpointType = EndpointType.ENDPOINT
    ) -> str:
        """Fetch the endpoint for a ledger DID.

        Args:
            did: The DID to look up on the ledger or in the cache
            endpoint_type: The type of the endpoint (default 'endpoint')
        """

    @abstractmethod
    async def get_all_endpoints_for_did(self, did: str) -> dict:
        """Fetch all endpoints for a ledger DID.

        Args:
            did: The DID to look up on the ledger or in the cache
        """

    async def _construct_attr_json(
        self,
        endpoint: str,
        endpoint_type: Optional[EndpointType] = None,
        all_exist_endpoints: Optional[dict] = None,
        routing_keys: Optional[List[str]] = None,
    ) -> str:
        """Create attr_json string.

        Args:
            all_exist_endpoints: Dictionary of all existing endpoints
            endpoint: The endpoint address
            endpoint_type: The type of the endpoint
            routing_keys: List of routing_keys if mediator is present
        """

        if not routing_keys:
            routing_keys = []

        if all_exist_endpoints:
            all_exist_endpoints[endpoint_type.indy] = endpoint
            all_exist_endpoints["routingKeys"] = routing_keys
            attr_json = json.dumps({"endpoint": all_exist_endpoints})

        else:
            endpoint_dict = {endpoint_type.indy: endpoint}
            endpoint_dict["routingKeys"] = routing_keys
            attr_json = json.dumps({"endpoint": endpoint_dict})

        return attr_json

    @abstractmethod
    async def update_endpoint_for_did(
        self,
        did: str,
        endpoint: str,
        endpoint_type: EndpointType = EndpointType.ENDPOINT,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
        routing_keys: Optional[List[str]] = None,
    ) -> bool:
        """Check and update the endpoint on the ledger.

        Args:
            did: The ledger DID
            endpoint: The endpoint address
            endpoint_type: The type of the endpoint (default 'endpoint')
            write_ledger: Flag to write the endpoint to the ledger
            endorser_did: Optional DID of the endorser
            routing_keys: List of routing_keys if mediator is present
        """

    @abstractmethod
    async def register_nym(
        self,
        did: str,
        verkey: str,
        alias: Optional[str] = None,
        role: Optional[str] = None,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> Tuple[bool, dict]:
        """Register a nym on the ledger.

        Args:
            did: DID to register on the ledger.
            verkey: The verification key of the keypair.
            alias: Human-friendly alias to assign to the DID.
            role: For permissioned ledgers, what role should the new DID have.
            write_ledger: Flag to write the nym to the ledger
            endorser_did: Optional DID of the endorser
        """

    @abstractmethod
    async def get_nym_role(self, did: str):
        """Return the role registered to input public DID on the ledger.

        Args:
            did: DID to register on the ledger.
        """

    @abstractmethod
    def nym_to_did(self, nym: str) -> str:
        """Format a nym with the ledger's DID prefix."""

    @abstractmethod
    async def rotate_public_did_keypair(self, next_seed: Optional[str] = None) -> None:
        """Rotate keypair for public DID: create new key, submit to ledger, update wallet.

        Args:
            next_seed: seed for incoming ed25519 keypair (default random)
        """

    @abstractmethod
    async def get_wallet_public_did(self) -> DIDInfo:
        """Fetch the public DID from the wallet."""

    @abstractmethod
    async def get_txn_author_agreement(self, reload: bool = False) -> dict:
        """Get the current transaction author agreement, fetching it if necessary."""

    @abstractmethod
    async def fetch_txn_author_agreement(self) -> dict:
        """Fetch the current AML and TAA from the ledger."""

    @abstractmethod
    async def accept_txn_author_agreement(
        self, taa_record: dict, mechanism: str, accept_time: Optional[int] = None
    ):
        """Save a new record recording the acceptance of the TAA."""

    @abstractmethod
    async def get_latest_txn_author_acceptance(self) -> dict:
        """Look up the latest TAA acceptance."""

    def taa_digest(self, version: str, text: str):
        """Generate the digest of a TAA record."""
        if not version or not text:
            raise ValueError("Bad input for TAA digest")
        taa_plaintext = version + text
        return sha256(taa_plaintext.encode("utf-8")).digest().hex()

    @abstractmethod
    async def txn_endorse(
        self,
        request_json: str,
        endorse_did: Optional[DIDInfo] = None,
    ) -> str:
        """Endorse (sign) the provided transaction."""

    @abstractmethod
    async def txn_submit(
        self,
        request_json: str,
        sign: bool,
        taa_accept: Optional[bool] = None,
        sign_did: DIDInfo = sentinel,
        write_ledger: bool = True,
    ) -> str:
        """Write the provided (signed and possibly endorsed) transaction to the ledger."""

    @abstractmethod
    async def fetch_schema_by_id(self, schema_id: str) -> dict:
        """Get schema from ledger.

        Args:
            schema_id: The schema id (or stringified sequence number) to retrieve

        Returns:
            Indy schema dict

        """

    @abstractmethod
    async def fetch_schema_by_seq_no(self, seq_no: int) -> dict:
        """Fetch a schema by its sequence number.

        Args:
            seq_no: schema ledger sequence number

        Returns:
            Indy schema dict

        """

    async def check_existing_schema(
        self,
        public_did: str,
        schema_name: str,
        schema_version: str,
        attribute_names: Sequence[str],
    ) -> Tuple[str, dict]:
        """Check if a schema has already been published."""
        fetch_schema_id = f"{public_did}:2:{schema_name}:{schema_version}"
        schema = await self.fetch_schema_by_id(fetch_schema_id)
        if schema:
            fetched_attrs = schema["attrNames"].copy()
            fetched_attrs.sort()
            cmp_attrs = list(attribute_names)
            cmp_attrs.sort()
            if fetched_attrs != cmp_attrs:
                raise LedgerTransactionError(
                    "Schema already exists on ledger, but attributes do not match: "
                    + f"{schema_name}:{schema_version} {fetched_attrs} != {cmp_attrs}"
                )
            return fetch_schema_id, schema

    async def create_and_send_schema(
        self,
        issuer: IndyIssuer,
        schema_name: str,
        schema_version: str,
        attribute_names: Sequence[str],
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> Tuple[str, dict]:
        """Send schema to ledger.

        Args:
            issuer: The issuer instance to use for schema creation
            schema_name: The schema name
            schema_version: The schema version
            attribute_names: A list of schema attributes
            write_ledger: Flag to write the schema to the ledger
            endorser_did: Optional DID of the endorser

        """

        public_info = await self.get_wallet_public_did()
        if not public_info:
            raise BadLedgerRequestError("Cannot publish schema without a public DID")

        if not bool(IndyDID.PATTERN.match(public_info.did)):
            raise BadLedgerRequestError(
                "Cannot publish schema when public DID is not an IndyDID"
            )

        schema_info = await self.check_existing_schema(
            public_info.did, schema_name, schema_version, attribute_names
        )
        if schema_info:
            LOGGER.warning("Schema already exists on ledger. Returning details.")
            schema_id, schema_def = schema_info
        else:
            if await self.is_ledger_read_only():
                raise LedgerError(
                    "Error cannot write schema when ledger is in read only mode, "
                    "or TAA is required and not accepted"
                )

            try:
                schema_id, schema_json = await issuer.create_schema(
                    public_info.did,
                    schema_name,
                    schema_version,
                    attribute_names,
                )
            except IndyIssuerError as err:
                raise LedgerError(err.message) from err
            schema_def = json.loads(schema_json)

            schema_req = await self._create_schema_request(
                public_info,
                schema_json,
                write_ledger=write_ledger,
                endorser_did=endorser_did,
            )

            try:
                resp = await self.txn_submit(
                    schema_req,
                    sign=True,
                    sign_did=public_info,
                    write_ledger=write_ledger,
                )

                if not write_ledger:
                    return schema_id, {"signed_txn": resp}

                try:
                    # parse sequence number out of response
                    seq_no = json.loads(resp)["result"]["txnMetadata"]["seqNo"]
                    schema_def["seqNo"] = seq_no
                except KeyError as err:
                    raise LedgerError(
                        "Failed to parse schema sequence number from ledger response"
                    ) from err
            except LedgerTransactionError as e:
                # Identify possible duplicate schema errors on indy-node < 1.9 and > 1.9
                if (
                    "can have one and only one SCHEMA with name" in e.message
                    or "UnauthorizedClientRequest" in e.message
                ):
                    # handle potential race condition if multiple agents are publishing
                    # the same schema simultaneously
                    schema_info = await self.check_existing_schema(
                        public_info.did, schema_name, schema_version, attribute_names
                    )
                    if schema_info:
                        LOGGER.warning(
                            "Schema already exists on ledger. Returning details."
                            " Error: %s",
                            e,
                        )
                        schema_id, schema_def = schema_info
                else:
                    raise

        return schema_id, schema_def

    @abstractmethod
    async def _create_schema_request(
        self,
        public_info: DIDInfo,
        schema_json: str,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ):
        """Create the ledger request for publishing a schema."""

    @abstractmethod
    async def _create_revoc_reg_def_request(
        self,
        public_info: DIDInfo,
        revoc_reg_def: dict,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ):
        """Create the ledger request for publishing a revocation registry definition."""

    @abstractmethod
    async def get_revoc_reg_def(self, revoc_reg_id: str) -> dict:
        """Look up a revocation registry definition by ID."""

    @abstractmethod
    async def send_revoc_reg_def(
        self,
        revoc_reg_def: dict,
        issuer_did: Optional[str] = None,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> dict:
        """Publish a revocation registry definition to the ledger."""

    @abstractmethod
    async def send_revoc_reg_entry(
        self,
        revoc_reg_id: str,
        revoc_def_type: str,
        revoc_reg_entry: dict,
        issuer_did: Optional[str] = None,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> dict:
        """Publish a revocation registry entry to the ledger."""

    async def create_and_send_credential_definition(
        self,
        issuer: IndyIssuer,
        schema_id: str,
        signature_type: Optional[str] = None,
        tag: Optional[str] = None,
        support_revocation: bool = False,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> Tuple[str, dict, bool]:
        """Send credential definition to ledger and store relevant key matter in wallet.

        Args:
            issuer: The issuer instance to use for credential definition creation
            schema_id: The schema id of the schema to create cred def for
            signature_type: The signature type to use on the credential definition
            tag: Optional tag to distinguish multiple credential definitions
            support_revocation: Optional flag to enable revocation for this cred def
            write_ledger: Flag to write the cred def to the ledger
            endorser_did: Optional DID of the endorser

        Returns:
            Tuple with cred def id, cred def structure, and whether it's novel

        """
        public_info = await self.get_wallet_public_did()
        if not public_info:
            raise BadLedgerRequestError(
                "Cannot publish credential definition without a public DID"
            )

        schema = await self.get_schema(schema_id)
        if not schema:
            raise LedgerError(f"Ledger {self.pool_name} has no schema {schema_id}")

        novel = False

        # check if cred def is on ledger already
        for test_tag in [tag] if tag else ["tag", DEFAULT_CRED_DEF_TAG]:
            credential_definition_id = issuer.make_credential_definition_id(
                strip_did_prefix(public_info.did), schema, signature_type, test_tag
            )
            ledger_cred_def = await self.fetch_credential_definition(
                credential_definition_id
            )
            if ledger_cred_def:
                LOGGER.warning(
                    "Credential definition %s already exists on ledger %s",
                    credential_definition_id,
                    self.pool_name,
                )

                try:
                    if not await issuer.credential_definition_in_wallet(
                        credential_definition_id
                    ):
                        raise LedgerError(
                            f"Credential definition {credential_definition_id} is on "
                            f"ledger {self.pool_name} but not in wallet "
                            f"{self.profile.name}"
                        )
                except IndyIssuerError as err:
                    raise LedgerError(err.message) from err

                credential_definition_json = json.dumps(ledger_cred_def)
                break
            else:  # no such cred def on ledger
                try:
                    if await issuer.credential_definition_in_wallet(
                        credential_definition_id
                    ):
                        raise LedgerError(
                            f"Credential definition {credential_definition_id} is in "
                            f"wallet {self.profile.name} but not on ledger "
                            f"{self.pool.name}"
                        )
                except IndyIssuerError as err:
                    raise LedgerError(err.message) from err

            # Cred def is neither on ledger nor in wallet: create and send it
            novel = True
            try:
                (
                    credential_definition_id,
                    credential_definition_json,
                ) = await issuer.create_and_store_credential_definition(
                    public_info.did,
                    schema,
                    signature_type,
                    tag,
                    support_revocation,
                )
            except IndyIssuerError as err:
                raise LedgerError(err.message) from err

            if await self.is_ledger_read_only():
                raise LedgerError(
                    "Error cannot write cred def when ledger is in read only mode, "
                    "or TAA is required and not accepted"
                )

            cred_def_req = await self._create_credential_definition_request(
                public_info,
                credential_definition_json,
                write_ledger=write_ledger,
                endorser_did=endorser_did,
            )

            resp = await self.txn_submit(
                cred_def_req, True, sign_did=public_info, write_ledger=write_ledger
            )
            if not write_ledger:
                return (credential_definition_id, {"signed_txn": resp}, novel)

        return (credential_definition_id, json.loads(credential_definition_json), novel)

    @abstractmethod
    async def _create_credential_definition_request(
        self,
        public_info: DIDInfo,
        credential_definition_json: str,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ):
        """Create the ledger request for publishing a credential definition."""

    @abstractmethod
    async def get_credential_definition(self, credential_definition_id: str) -> dict:
        """Get a credential definition from the cache if available, otherwise the ledger.

        Args:
            credential_definition_id: The schema id of the schema to fetch cred def for

        """

    @abstractmethod
    async def get_revoc_reg_delta(
        self, revoc_reg_id: str, timestamp_from=0, timestamp_to=None
    ) -> Tuple[dict, int]:
        """Look up a revocation registry delta by ID."""

    @abstractmethod
    async def get_schema(self, schema_id: str) -> dict:
        """Get a schema from the cache if available, otherwise fetch from the ledger.

        Args:
            schema_id: The schema id (or stringified sequence number) to retrieve

        """

    @abstractmethod
    async def get_revoc_reg_entry(
        self, revoc_reg_id: str, timestamp: int
    ) -> Tuple[dict, int]:
        """Get revocation registry entry by revocation registry ID and timestamp."""

    async def check_existing_schema_anoncreds(
        self,
        schema_id: str,
        attribute_names: Sequence[str],
    ) -> Tuple[str, dict]:
        """Check if a schema has already been published."""
        schema = await self.fetch_schema_by_id(schema_id)
        if schema:
            fetched_attrs = schema["attrNames"].copy()
            if set(fetched_attrs) != set(attribute_names):
                raise LedgerTransactionError(
                    "Schema already exists on ledger, but attributes do not match: "
                    + f"{schema_id} {fetched_attrs} != {attribute_names}"
                )
            return schema_id, schema

    async def send_schema_anoncreds(
        self,
        schema_id: str,
        schema_def: dict,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> Tuple[str, dict]:
        """Send schema to the ledger.

        This method sends a schema to the ledger for publication.

        Args:
            schema_id (str): The ID of the schema.
            schema_def (dict): The definition of the schema.
            write_ledger (bool, optional): Whether to write the schema to the ledger.
                Defaults to True.
            endorser_did (str, optional): The DID of the endorser. Defaults to None.

        Returns:
            Tuple[str, dict]: A tuple containing the schema ID and the schema definition.

        Raises:
            BadLedgerRequestError: If there is no public DID available for publishing the
                schema.
            BadLedgerRequestError: If the public DID is not an IndyDID.
            LedgerError: If the ledger is in read-only mode or if the TAA is required and
                not accepted.
            LedgerError: If the ledger is in read-only mode.
            LedgerError: If parsing the schema sequence number from the ledger response
                fails.
            LedgerObjectAlreadyExistsError: If the schema already exists on the ledger.

        """
        from acapy_agent.anoncreds.default.legacy_indy.registry import LegacyIndyRegistry

        public_info = await self.get_wallet_public_did()
        if not public_info:
            raise BadLedgerRequestError("Cannot publish schema without a public DID")

        if not bool(IndyDID.PATTERN.match(public_info.did)):
            raise BadLedgerRequestError(
                "Cannot publish schema when public DID is not an IndyDID"
            )

        schema_info = await self.check_existing_schema_anoncreds(
            schema_id, schema_def["attrNames"]
        )

        if schema_info:
            LOGGER.warning("Schema already exists on ledger. Returning details.")
            schema_id, schema_def = schema_info
        else:
            if await self.is_ledger_read_only():
                raise LedgerError(
                    "Error cannot write schema when ledger is in read only mode, "
                    "or TAA is required and not accepted"
                )

        if await self.is_ledger_read_only():
            raise LedgerError(
                "Error cannot write schema when ledger is in read only mode"
            )

        schema_req = await self._create_schema_request(
            public_info,
            json.dumps(schema_def),
            write_ledger=write_ledger,
            endorser_did=endorser_did,
        )

        try:
            legacy_indy_registry = LegacyIndyRegistry()
            resp = await legacy_indy_registry.txn_submit(
                self,
                schema_req,
                sign=True,
                sign_did=public_info,
                write_ledger=write_ledger,
            )

            if not write_ledger:
                return schema_id, {"signed_txn": resp}

            try:
                # parse sequence number out of response
                seq_no = json.loads(resp)["result"]["txnMetadata"]["seqNo"]
                return seq_no
            except KeyError as err:
                raise LedgerError(
                    "Failed to parse schema sequence number from ledger response"
                ) from err
        except LedgerTransactionError as e:
            # Identify possible duplicate schema errors on indy-node < 1.9 and > 1.9
            if (
                "can have one and only one SCHEMA with name" in e.message
                or "UnauthorizedClientRequest" in e.message
            ):
                # handle potential race condition if multiple agents are publishing
                # the same schema simultaneously
                schema_info = await self.check_existing_schema_anoncreds(
                    schema_id, schema_def["attrNames"]
                )
                if schema_info:
                    LOGGER.warning(
                        "Schema already exists on ledger. Returning details. Error: %s",
                        e,
                    )
                    raise LedgerObjectAlreadyExistsError(
                        f"Schema already exists on ledger (Error: {e})", *schema_info
                    )
                else:
                    raise
            else:
                raise

    async def send_credential_definition_anoncreds(
        self,
        schema_id: str,
        cred_def_id: str,
        cred_def: dict,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> Tuple[str, dict, bool]:
        """Send credential definition to ledger and store relevant key matter in wallet.

        Args:
            schema_id (str): The schema id of the schema to create credential definition
                for.
            cred_def_id (str): The credential definition id.
            cred_def (dict): The credential definition structure.
            write_ledger (bool, optional): Flag indicating whether to write the
                credential definition to the ledger. Defaults to True.
            endorser_did (str, optional): The DID of the endorser. Defaults to None.

        Returns:
            Tuple[str, dict, bool]: A tuple containing the credential definition id,
                credential definition structure, and a boolean indicating whether it's
                novel.

        Raises:
            BadLedgerRequestError: If there is no public DID available to publish the
                credential definition.
            LedgerError: If the ledger does not have the specified schema.
            LedgerObjectAlreadyExistsError: If the credential definition already exists
                in the wallet and on the ledger.
            LedgerError: If the ledger is in read-only mode and cannot write the
                credential definition.

        """
        public_info = await self.get_wallet_public_did()
        if not public_info:
            raise BadLedgerRequestError(
                "Cannot publish credential definition without a public DID"
            )

        schema = await self.get_schema(schema_id)
        if not schema:
            raise LedgerError(f"Ledger {self.pool.name} has no schema {schema_id}")

        # check if cred def is on ledger already
        ledger_cred_def = await self.fetch_credential_definition(cred_def_id)
        if ledger_cred_def:
            credential_definition_json = json.dumps(ledger_cred_def)
            raise LedgerObjectAlreadyExistsError(
                f"Credential definition with id {cred_def_id} "
                "already exists in wallet and on ledger.",
                cred_def_id,
                credential_definition_json,
            )

        if await self.is_ledger_read_only():
            raise LedgerError(
                "Error cannot write cred def when ledger is in read only mode"
            )

        cred_def_req = await self._create_credential_definition_request(
            public_info,
            json.dumps(cred_def),
            write_ledger=write_ledger,
            endorser_did=endorser_did,
        )

        resp = await self.txn_submit(
            cred_def_req, True, sign_did=public_info, write_ledger=write_ledger
        )

        if not write_ledger:
            return (cred_def_id, {"signed_txn": resp})

        seq_no = json.loads(resp)["result"]["txnMetadata"]["seqNo"]
        return seq_no


class Role(Enum):
    """Enum for indy roles."""

    STEWARD = (2,)
    TRUSTEE = (0,)
    ENDORSER = (101,)
    NETWORK_MONITOR = (201,)
    USER = (None, "")  # in case reading from file, default empty "" or None for USER
    ROLE_REMOVE = ("",)  # but indy-sdk uses "" to identify a role in reset

    @staticmethod
    def get(token: Union[str, int] = None) -> "Role":
        """Return enum instance corresponding to input token.

        Args:
            token: token identifying role to indy-sdk:
                "STEWARD", "TRUSTEE", "ENDORSER", "" or None
        """
        if token is None:
            return Role.USER

        for role in Role:
            if role == Role.ROLE_REMOVE:
                continue  # not a sensible role to parse from any configuration
            if isinstance(token, int) and token in role.value:
                return role
            if str(token).upper() == role.name or token in (str(v) for v in role.value):
                return role

        return None

    def to_indy_num_str(self) -> str:
        """Return (typically, numeric) string value that indy-sdk associates with role.

        Recall that None signifies USER and "" signifies a role undergoing reset.
        """

        return str(self.value[0]) if isinstance(self.value[0], int) else self.value[0]

    def token(self) -> str:
        """Return token identifying role to indy-sdk."""

        return self.value[0] if self in (Role.USER, Role.ROLE_REMOVE) else self.name



"""anoncreds-rs issuer implementation."""

import asyncio
import logging
from time import time
from typing import Optional, Sequence

from anoncreds import (
    AnoncredsError,
    Credential,
    CredentialDefinition,
    CredentialDefinitionPrivate,
    CredentialOffer,
    KeyCorrectnessProof,
    Schema,
    W3cCredential,
)
from aries_askar import AskarError

from ..askar.profile_anon import AskarAnonCredsProfile, AskarAnonCredsProfileSession
from ..core.error import BaseError
from ..core.event_bus import Event, EventBus
from ..core.profile import Profile
from ..protocols.endorse_transaction.v1_0.util import is_author_role
from .base import AnonCredsSchemaAlreadyExists, BaseAnonCredsError
from .error_messages import ANONCREDS_PROFILE_REQUIRED_MSG
from .events import CredDefFinishedEvent
from .models.credential_definition import CredDef, CredDefResult
from .models.schema import AnonCredsSchema, SchemaResult, SchemaState
from .registry import AnonCredsRegistry

LOGGER = logging.getLogger(__name__)

DEFAULT_CRED_DEF_TAG = "default"
DEFAULT_SIGNATURE_TYPE = "CL"
DEFAULT_MAX_CRED_NUM = 1000
CATEGORY_SCHEMA = "schema"
CATEGORY_CRED_DEF = "credential_def"
CATEGORY_CRED_DEF_PRIVATE = "credential_def_private"
CATEGORY_CRED_DEF_KEY_PROOF = "credential_def_key_proof"
STATE_FINISHED = "finished"

EVENT_PREFIX = "acapy::anoncreds::"
EVENT_SCHEMA = EVENT_PREFIX + CATEGORY_SCHEMA
EVENT_CRED_DEF = EVENT_PREFIX + CATEGORY_CRED_DEF
EVENT_FINISHED_SUFFIX = "::" + STATE_FINISHED


class AnonCredsIssuerError(BaseError):
    """Generic issuer error."""


class AnonCredsIssuer:
    """AnonCreds issuer class.

    This class provides methods for creating and registering AnonCreds objects
    needed to issue credentials. It also provides methods for storing and
    retrieving local representations of these objects from the wallet.

    A general pattern is followed when creating and registering objects:

    1. Create the object locally
    2. Register the object with the anoncreds registry
    3. Store the object in the wallet

    The wallet storage is used to keep track of the state of the object.

    If the object is fully registered immediately after sending to the registry
    (state of `finished`), the object is saved to the wallet with an id
    matching the id returned from the registry.

    If the object is not fully registered but pending (state of `wait`), the
    object is saved to the wallet with an id matching the job id returned from
    the registry.

    If the object fails to register (state of `failed`), the object is saved to
    the wallet with an id matching the job id returned from the registry.

    When an object finishes registration after being in a pending state (moving
    from state `wait` to state `finished`), the wallet entry matching the job id
    is removed and an entry matching the registered id is added.
    """

    def __init__(self, profile: Profile):
        """Initialize an AnonCredsIssuer instance.

        Args:
            profile: The active profile instance

        """
        self._profile = profile

    @property
    def profile(self) -> AskarAnonCredsProfile:
        """Accessor for the profile instance."""
        if not isinstance(self._profile, AskarAnonCredsProfile):
            raise ValueError(ANONCREDS_PROFILE_REQUIRED_MSG)

        return self._profile

    async def notify(self, event: Event) -> None:
        """Accessor for the event bus instance."""
        event_bus = self.profile.inject(EventBus)
        await event_bus.notify(self._profile, event)

    async def _finish_registration(
        self,
        txn: AskarAnonCredsProfileSession,
        category: str,
        job_id: str,
        registered_id: str,
    ):
        entry = await txn.handle.fetch(
            category,
            job_id,
            for_update=True,
        )
        if not entry:
            raise AnonCredsIssuerError(
                f"{category} with job id {job_id} could not be found"
            )

        tags = entry.tags
        tags["state"] = STATE_FINISHED
        await txn.handle.insert(
            category,
            registered_id,
            value=entry.value,
            tags=tags,
        )
        await txn.handle.remove(category, job_id)
        return entry

    async def store_schema(
        self,
        result: SchemaResult,
    ) -> None:
        """Store schema after reaching finished state."""
        identifier = result.job_id or result.schema_state.schema_id
        if not identifier:
            raise ValueError("Schema id or job id must be set")

        try:
            async with self.profile.session() as session:
                await session.handle.insert(
                    CATEGORY_SCHEMA,
                    identifier,
                    result.schema_state.schema.to_json(),
                    {
                        "name": result.schema_state.schema.name,
                        "version": result.schema_state.schema.version,
                        "issuer_id": result.schema_state.schema.issuer_id,
                        "state": result.schema_state.state,
                    },
                )
        except AskarError as err:
            raise AnonCredsIssuerError("Error storing schema") from err

    async def create_and_register_schema(
        self,
        issuer_id: str,
        name: str,
        version: str,
        attr_names: Sequence[str],
        options: Optional[dict] = None,
    ) -> SchemaResult:
        """Create a new credential schema and store it in the wallet.

        Args:
            issuer_id (str): The DID issuing the credential definition.
            name (str): The schema name.
            version (str): The schema version.
            attr_names (Sequence[str]): A sequence of schema attribute names.
            options (Optional[dict]): Additional options for schema creation.
                Defaults to None.

        Returns:
            SchemaResult: An instance of SchemaResult.

        Raises:
            AnonCredsSchemaAlreadyExists: If a similar schema already exists in
                the records.

        """
        options = options or {}
        # Check if record of a similar schema already exists in our records
        async with self.profile.session() as session:
            # TODO scan?
            schemas = await session.handle.fetch_all(
                CATEGORY_SCHEMA,
                {
                    "name": name,
                    "version": version,
                    "issuer_id": issuer_id,
                },
                limit=1,
            )
            if schemas:
                raise AnonCredsSchemaAlreadyExists(
                    f"Schema with {name}: {version} already exists for {issuer_id}",
                    schemas[0].name,
                    AnonCredsSchema.deserialize(schemas[0].value_json),
                )

        schema = Schema.create(name, version, issuer_id, attr_names)
        try:
            anoncreds_registry = self.profile.inject(AnonCredsRegistry)
            schema_result = await anoncreds_registry.register_schema(
                self.profile,
                AnonCredsSchema.from_native(schema),
                options,
            )

            await self.store_schema(schema_result)
            return schema_result

        except AnonCredsSchemaAlreadyExists as err:
            # If we find that we've previously written a schema that looks like
            # this one before but that schema is not in our wallet, add it to
            # the wallet so we can return from our get schema calls
            await self.store_schema(
                SchemaResult(
                    job_id=None,
                    schema_state=SchemaState(
                        state=SchemaState.STATE_FINISHED,
                        schema_id=err.schema_id,
                        schema=err.schema,
                    ),
                )
            )
            raise AnonCredsIssuerError(
                "Schema already exists but was not in wallet; stored in wallet"
            ) from err
        except (AnoncredsError, BaseAnonCredsError) as err:
            raise AnonCredsIssuerError("Error creating schema") from err

    async def finish_schema(self, job_id: str, schema_id: str) -> None:
        """Mark a schema as finished."""
        async with self.profile.transaction() as txn:
            await self._finish_registration(txn, CATEGORY_SCHEMA, job_id, schema_id)
            await txn.commit()

    async def get_created_schemas(
        self,
        name: Optional[str] = None,
        version: Optional[str] = None,
        issuer_id: Optional[str] = None,
    ) -> Sequence[str]:
        """Retrieve IDs of schemas previously created."""
        async with self.profile.session() as session:
            # TODO limit? scan?
            schemas = await session.handle.fetch_all(
                CATEGORY_SCHEMA,
                {
                    key: value
                    for key, value in {
                        "name": name,
                        "version": version,
                        "issuer_id": issuer_id,
                        "state": STATE_FINISHED,
                    }.items()
                    if value is not None
                },
            )
        # entry.name was stored as the schema's ID
        return [entry.name for entry in schemas]

    async def credential_definition_in_wallet(
        self, credential_definition_id: str
    ) -> bool:
        """Check whether a given credential definition ID is present in the wallet.

        Args:
            credential_definition_id: The credential definition ID to check
        """
        try:
            async with self.profile.session() as session:
                return (
                    await session.handle.fetch(
                        CATEGORY_CRED_DEF_PRIVATE, credential_definition_id
                    )
                ) is not None
        except AskarError as err:
            raise AnonCredsIssuerError(
                "Error checking for credential definition"
            ) from err

    async def create_and_register_credential_definition(
        self,
        issuer_id: str,
        schema_id: str,
        tag: Optional[str] = None,
        signature_type: Optional[str] = None,
        options: Optional[dict] = None,
    ) -> CredDefResult:
        """Create a new credential definition and store it in the wallet.

        Args:
            issuer_id: the ID of the issuer creating the credential definition
            schema_id: the schema ID for the credential definition
            tag: the tag to use for the credential definition
            signature_type: the signature type to use for the credential definition
            options: any additional options to use when creating the credential definition

        Returns:
            CredDefResult: the result of the credential definition creation

        """
        options = options or {}
        support_revocation_option = options.get("support_revocation")

        if support_revocation_option is None:
            # Support revocation not set - Default to auto-create rev reg if author role
            is_author = is_author_role(self.profile)
            auto_create_rev_reg = self.profile.settings.get(
                "endorser.auto_create_rev_reg", False
            )

            support_revocation = bool(is_author and auto_create_rev_reg)
        else:
            # If support_revocation is explicitly set, use that value
            if not isinstance(support_revocation_option, bool):
                raise ValueError("support_revocation must be a boolean")

            support_revocation = support_revocation_option

        max_cred_num = options.get("revocation_registry_size", DEFAULT_MAX_CRED_NUM)
        if not isinstance(max_cred_num, int):
            raise ValueError("max_cred_num must be an integer")

        # Don't allow revocable cred def to be created without tails server base url
        if not self.profile.settings.get("tails_server_base_url") and support_revocation:
            raise AnonCredsIssuerError(
                "tails_server_base_url not configured. "
                "Can't create revocable credential definition."
            )

        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        schema_result = await anoncreds_registry.get_schema(self.profile, schema_id)

        # Create the cred def
        (
            cred_def,
            cred_def_private,
            key_proof,
        ) = await asyncio.get_event_loop().run_in_executor(
            None,
            lambda: CredentialDefinition.create(
                schema_id=schema_id,
                schema=schema_result.schema.serialize(),
                issuer_id=issuer_id,
                tag=tag or DEFAULT_CRED_DEF_TAG,
                signature_type=signature_type or DEFAULT_SIGNATURE_TYPE,
                support_revocation=support_revocation,
            ),
        )

        try:
            cred_def_result = await anoncreds_registry.register_credential_definition(
                profile=self.profile,
                schema=schema_result,
                credential_definition=CredDef.from_native(cred_def),
                options=options,
            )

            await self.store_credential_definition(
                schema_result=schema_result,
                cred_def_result=cred_def_result,
                cred_def_private=cred_def_private,
                key_proof=key_proof,
                support_revocation=support_revocation,
                max_cred_num=max_cred_num,
                options=options,
            )

            return cred_def_result
        except (AnoncredsError, BaseAnonCredsError) as err:
            raise AnonCredsIssuerError("Error creating credential definition") from err

    async def store_credential_definition(
        self,
        schema_result: SchemaResult,
        cred_def_result: CredDefResult,
        cred_def_private: CredentialDefinitionPrivate,
        key_proof: KeyCorrectnessProof,
        support_revocation: bool,
        max_cred_num: int,
        options: Optional[dict] = None,
    ) -> None:
        """Store the cred def and it's components in the wallet."""
        options = options or {}
        identifier = (
            cred_def_result.job_id
            or cred_def_result.credential_definition_state.credential_definition_id
        )

        if not identifier:
            raise AnonCredsIssuerError("cred def id or job id required")

        try:
            async with self.profile.transaction() as txn:
                await txn.handle.insert(
                    CATEGORY_CRED_DEF,
                    identifier,
                    cred_def_result.credential_definition_state.credential_definition.to_json(),
                    tags={
                        "schema_id": schema_result.schema_id,
                        "schema_issuer_id": schema_result.schema.issuer_id,
                        "issuer_id": cred_def_result.credential_definition_state.credential_definition.issuer_id,  # noqa: E501
                        "schema_name": schema_result.schema.name,
                        "schema_version": schema_result.schema.version,
                        "state": cred_def_result.credential_definition_state.state,
                        "epoch": str(int(time())),
                        # TODO We need to keep track of these but tags probably
                        # isn't ideal. This suggests that a full record object
                        # is necessary for non-private values
                        "support_revocation": str(support_revocation),
                        "max_cred_num": str(max_cred_num),
                    },
                )
                await txn.handle.insert(
                    CATEGORY_CRED_DEF_PRIVATE,
                    identifier,
                    cred_def_private.to_json_buffer(),
                )
                await txn.handle.insert(
                    CATEGORY_CRED_DEF_KEY_PROOF, identifier, key_proof.to_json_buffer()
                )
                await txn.commit()
            if cred_def_result.credential_definition_state.state == STATE_FINISHED:
                await self.notify(
                    CredDefFinishedEvent.with_payload(
                        schema_id=schema_result.schema_id,
                        cred_def_id=identifier,
                        issuer_id=cred_def_result.credential_definition_state.credential_definition.issuer_id,
                        support_revocation=support_revocation,
                        max_cred_num=max_cred_num,
                        options=options,
                    )
                )
        except AskarError as err:
            raise AnonCredsIssuerError("Error storing credential definition") from err

    async def finish_cred_def(
        self, job_id: str, cred_def_id: str, options: Optional[dict] = None
    ) -> None:
        """Finish a cred def."""
        async with self.profile.transaction() as txn:
            entry = await self._finish_registration(
                txn, CATEGORY_CRED_DEF, job_id, cred_def_id
            )
            cred_def = CredDef.from_json(entry.value)
            support_revocation = entry.tags["support_revocation"] == "True"
            max_cred_num = int(entry.tags["max_cred_num"])

            await self._finish_registration(
                txn, CATEGORY_CRED_DEF_PRIVATE, job_id, cred_def_id
            )
            await self._finish_registration(
                txn, CATEGORY_CRED_DEF_KEY_PROOF, job_id, cred_def_id
            )
            await txn.commit()

        await self.notify(
            CredDefFinishedEvent.with_payload(
                schema_id=cred_def.schema_id,
                cred_def_id=cred_def_id,
                issuer_id=cred_def.issuer_id,
                support_revocation=support_revocation,
                max_cred_num=max_cred_num,
                options=options,
            )
        )

    async def get_created_credential_definitions(
        self,
        issuer_id: Optional[str] = None,
        schema_issuer_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_name: Optional[str] = None,
        schema_version: Optional[str] = None,
        epoch: Optional[str] = None,
    ) -> Sequence[str]:
        """Retrieve IDs of credential definitions previously created."""
        async with self.profile.session() as session:
            # TODO limit? scan?
            credential_definition_entries = await session.handle.fetch_all(
                CATEGORY_CRED_DEF,
                {
                    key: value
                    for key, value in {
                        "issuer_id": issuer_id,
                        "schema_issuer_id": schema_issuer_id,
                        "schema_id": schema_id,
                        "schema_name": schema_name,
                        "schema_version": schema_version,
                        "epoch": epoch,
                        "state": STATE_FINISHED,
                    }.items()
                    if value is not None
                },
            )
        # entry.name is cred def id when state == finished
        return [entry.name for entry in credential_definition_entries]

    async def match_created_credential_definitions(
        self,
        cred_def_id: Optional[str] = None,
        issuer_id: Optional[str] = None,
        schema_issuer_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_name: Optional[str] = None,
        schema_version: Optional[str] = None,
        epoch: Optional[str] = None,
    ) -> Optional[str]:
        """Return cred def id of most recent matching cred def."""
        async with self.profile.session() as session:
            # TODO limit? scan?
            if cred_def_id:
                cred_def_entry = await session.handle.fetch(
                    CATEGORY_CRED_DEF, cred_def_id
                )
            else:
                credential_definition_entries = await session.handle.fetch_all(
                    CATEGORY_CRED_DEF,
                    {
                        key: value
                        for key, value in {
                            "issuer_id": issuer_id,
                            "schema_issuer_id": schema_issuer_id,
                            "schema_id": schema_id,
                            "schema_name": schema_name,
                            "schema_version": schema_version,
                            "state": STATE_FINISHED,
                            "epoch": epoch,
                        }.items()
                        if value is not None
                    },
                )
                cred_def_entry = max(
                    list(credential_definition_entries),
                    key=lambda r: int(r.tags["epoch"]),
                )

        if cred_def_entry:
            return cred_def_entry.name

        return None

    async def cred_def_supports_revocation(self, cred_def_id: str) -> bool:
        """Return whether a credential definition supports revocation."""
        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        cred_def_result = await anoncreds_registry.get_credential_definition(
            self.profile, cred_def_id
        )
        return cred_def_result.credential_definition.value.revocation is not None

    async def create_credential_offer(self, credential_definition_id: str) -> str:
        """Create a credential offer for the given credential definition id.

        Args:
            credential_definition_id: The credential definition to create an offer for

        Returns:
            The new credential offer

        """
        try:
            async with self.profile.session() as session:
                cred_def = await session.handle.fetch(
                    CATEGORY_CRED_DEF, credential_definition_id
                )
                key_proof = await session.handle.fetch(
                    CATEGORY_CRED_DEF_KEY_PROOF, credential_definition_id
                )
        except AskarError as err:
            raise AnonCredsIssuerError("Error retrieving credential definition") from err
        if not cred_def or not key_proof:
            raise AnonCredsIssuerError(
                "Credential definition not found for credential offer"
            )
        try:
            # The tag holds the full name of the schema,
            # as opposed to just the sequence number
            schema_id = cred_def.tags.get("schema_id")
            cred_def = CredentialDefinition.load(cred_def.raw_value)

            credential_offer = CredentialOffer.create(
                schema_id or cred_def.schema_id,
                credential_definition_id,
                key_proof.raw_value,
            )
        except AnoncredsError as err:
            raise AnonCredsIssuerError("Error creating credential offer") from err

        return credential_offer.to_json()

    async def create_credential(
        self,
        credential_offer: dict,
        credential_request: dict,
        credential_values: dict,
    ) -> str:
        """Create Credential."""
        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        schema_id = credential_offer["schema_id"]
        schema_result = await anoncreds_registry.get_schema(self.profile, schema_id)
        cred_def_id = credential_offer["cred_def_id"]
        schema_attributes = schema_result.schema_value.attr_names

        try:
            async with self.profile.session() as session:
                cred_def = await session.handle.fetch(CATEGORY_CRED_DEF, cred_def_id)
                cred_def_private = await session.handle.fetch(
                    CATEGORY_CRED_DEF_PRIVATE, cred_def_id
                )
        except AskarError as err:
            raise AnonCredsIssuerError("Error retrieving credential definition") from err

        if not cred_def or not cred_def_private:
            raise AnonCredsIssuerError(
                "Credential definition not found for credential issuance"
            )

        raw_values = {}
        for attribute in schema_attributes:
            # Ensure every attribute present in schema to be set.
            # Extraneous attribute names are ignored.
            try:
                credential_value = credential_values[attribute]
            except KeyError:
                raise AnonCredsIssuerError(
                    "Provided credential values are missing a value "
                    f"for the schema attribute '{attribute}'"
                )

            raw_values[attribute] = str(credential_value)

        try:
            credential = await asyncio.get_event_loop().run_in_executor(
                None,
                lambda: Credential.create(
                    cred_def.raw_value,
                    cred_def_private.raw_value,
                    credential_offer,
                    credential_request,
                    raw_values,
                ),
            )
        except AnoncredsError as err:
            raise AnonCredsIssuerError("Error creating credential") from err

        return credential.to_json()

    async def create_credential_w3c(
        self,
        credential_offer: dict,
        credential_request: dict,
        credential_values: dict,
    ) -> str:
        """Create Credential."""
        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        schema_id = credential_offer["schema_id"]
        schema_result = await anoncreds_registry.get_schema(self.profile, schema_id)
        cred_def_id = credential_offer["cred_def_id"]
        schema_attributes = schema_result.schema_value.attr_names

        try:
            async with self.profile.session() as session:
                cred_def = await session.handle.fetch(CATEGORY_CRED_DEF, cred_def_id)
                cred_def_private = await session.handle.fetch(
                    CATEGORY_CRED_DEF_PRIVATE, cred_def_id
                )
        except AskarError as err:
            raise AnonCredsIssuerError("Error retrieving credential definition") from err

        if not cred_def or not cred_def_private:
            raise AnonCredsIssuerError(
                "Credential definition not found for credential issuance"
            )

        raw_values = {}
        for attribute in schema_attributes:
            # Ensure every attribute present in schema to be set.
            # Extraneous attribute names are ignored.
            try:
                credential_value = credential_values[attribute]
            except KeyError:
                raise AnonCredsIssuerError(
                    "Provided credential values are missing a value "
                    f"for the schema attribute '{attribute}'"
                )

            raw_values[attribute] = str(credential_value)

        try:
            credential = await asyncio.get_event_loop().run_in_executor(
                None,
                lambda: W3cCredential.create(
                    cred_def.raw_value,
                    cred_def_private.raw_value,
                    credential_offer,
                    credential_request,
                    raw_values,
                ),
            )
        except AnoncredsError as err:
            raise AnonCredsIssuerError("Error creating credential") from err

        return credential.to_json()



"""Handler for mediate-deny message."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageNotFoundError
from ..manager import MediationManager
from ..messages.mediate_deny import MediationDeny
from ..models.mediation_record import MediationRecord


class MediationDenyHandler(BaseHandler):
    """Handler for mediate-deny message."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle mediate-deny message."""
        self._logger.debug("%s called with context %s", self.__class__.__name__, context)
        assert isinstance(context.message, MediationDeny)

        if not context.connection_ready:
            raise HandlerException("Received mediation deny from inactive connection")

        profile = context.profile
        mgr = MediationManager(profile)
        try:
            async with profile.session() as session:
                record = await MediationRecord.retrieve_by_connection_id(
                    session, context.connection_record.connection_id
                )
            await mgr.request_denied(record, context.message)
        except StorageNotFoundError as err:
            raise HandlerException(
                "Received mediation grant from connection from which mediation "
                "has not been requested"
            ) from err



"""The Dispatcher.

The dispatcher is responsible for coordinating data flow between handlers, providing
lifecycle hook callbacks storing state for message threads, etc.
"""

import asyncio
import logging
import os
import warnings
import weakref
from typing import Callable, Coroutine, Optional, Union

from aiohttp.web import HTTPException

from ..connections.base_manager import BaseConnectionManager
from ..connections.models.conn_record import ConnRecord
from ..core.profile import Profile
from ..messaging.agent_message import AgentMessage
from ..messaging.base_message import BaseMessage, DIDCommVersion
from ..messaging.error import MessageParseError
from ..messaging.models.base import BaseModelError
from ..messaging.request_context import RequestContext
from ..messaging.responder import SKIP_ACTIVE_CONN_CHECK_MSG_TYPES, BaseResponder
from ..messaging.util import datetime_now
from ..messaging.v2_agent_message import V2AgentMessage
from ..protocols.problem_report.v1_0.message import ProblemReport
from ..transport.inbound.message import InboundMessage
from ..transport.outbound.message import OutboundMessage
from ..transport.outbound.status import OutboundSendStatus
from ..utils.classloader import DeferLoad
from ..utils.stats import Collector
from ..utils.task_queue import CompletedTask, PendingTask, TaskQueue
from ..utils.tracing import get_timer, trace_event
from .error import ProtocolMinorVersionNotSupported
from .protocol_registry import ProtocolRegistry


class ProblemReportParseError(MessageParseError):
    """Error to raise on failure to parse problem-report message."""


class Dispatcher:
    """Dispatcher class.

    Class responsible for dispatching messages to message handlers and responding
    to other agents.
    """

    def __init__(self, profile: Profile):
        """Initialize an instance of Dispatcher."""
        self.collector: Optional[Collector] = None
        self.profile = profile
        self.task_queue: Optional[TaskQueue] = None
        self.logger: logging.Logger = logging.getLogger(__name__)

    async def setup(self):
        """Perform async instance setup."""
        self.collector = self.profile.inject_or(Collector)
        max_active = int(os.getenv("DISPATCHER_MAX_ACTIVE", 50))
        self.task_queue = TaskQueue(
            max_active=max_active, timed=bool(self.collector), trace_fn=self.log_task
        )

    def put_task(
        self,
        coro: Coroutine,
        complete: Optional[Callable] = None,
        ident: Optional[str] = None,
    ) -> PendingTask:
        """Run a task in the task queue, potentially blocking other handlers."""
        return self.task_queue.put(coro, complete, ident)

    def run_task(
        self,
        coro: Coroutine,
        complete: Optional[Callable] = None,
        ident: Optional[str] = None,
    ) -> asyncio.Task:
        """Run a task in the task queue, potentially blocking other handlers."""
        return self.task_queue.run(coro, complete, ident)

    def log_task(self, task: CompletedTask):
        """Log a completed task using the stats collector."""
        if task.exc_info and not issubclass(task.exc_info[0], HTTPException):
            # skip errors intentionally returned to HTTP clients
            self.logger.exception(
                "Handler error: %s", task.ident or "", exc_info=task.exc_info
            )
        if self.collector:
            timing = task.timing
            if "queued" in timing:
                self.collector.log(
                    "Dispatcher:queued", timing["unqueued"] - timing["queued"]
                )
            if task.ident:
                self.collector.log(task.ident, timing["ended"] - timing["started"])

    def queue_message(
        self,
        profile: Profile,
        inbound_message: InboundMessage,
        send_outbound: Coroutine,
        complete: Optional[Callable] = None,
    ) -> PendingTask:
        """Add a message to the processing queue for handling.

        Args:
            profile: The profile associated with the inbound message
            inbound_message: The inbound message instance
            send_outbound: Async function to send outbound messages
            complete: Function to call when the handler has completed

        Returns:
            A pending task instance resolving to the handler task

        """

        if (
            self.profile.settings.get("experiment.didcomm_v2")
            and inbound_message.receipt.didcomm_version == DIDCommVersion.v2
        ):
            handle = self.handle_v2_message(profile, inbound_message, send_outbound)
        else:
            handle = self.handle_v1_message(profile, inbound_message, send_outbound)

        return self.put_task(
            handle,
            complete,
        )

    async def handle_v2_message(
        self,
        profile: Profile,
        inbound_message: InboundMessage,
        send_outbound: Coroutine,
    ):
        """Handle a DIDComm V2 message."""

        # send a DCV2 Problem Report here for testing, and to punt procotol handling down
        # the road a bit
        context = RequestContext(profile)
        context.message_receipt = inbound_message.receipt
        responder = DispatcherResponder(
            context,
            inbound_message,
            send_outbound,
            reply_session_id=inbound_message.session_id,
            reply_to_verkey=inbound_message.receipt.sender_verkey,
        )

        context.injector.bind_instance(BaseResponder, responder)
        error_result = V2AgentMessage(
            message={
                "type": "https://didcomm.org/report-problem/2.0/problem-report",
                "body": {
                    "comment": "No Handlers Found",
                    "code": "e.p.msg.not-found",
                },
            }
        )
        if inbound_message.receipt.thread_id:
            error_result.message["pthid"] = inbound_message.receipt.thread_id
        await responder.send_reply(error_result)

    async def handle_v1_message(
        self,
        profile: Profile,
        inbound_message: InboundMessage,
        send_outbound: Coroutine,
    ):
        """Configure responder and message context and invoke the message handler.

        Args:
            profile: The profile associated with the inbound message
            inbound_message: The inbound message instance
            send_outbound: Async function to send outbound messages

        Returns:
            The response from the handler

        """
        r_time = get_timer()

        error_result = None
        version_warning = None
        message = None
        try:
            message = await self.make_message(profile, inbound_message.payload)
        except ProblemReportParseError:
            pass  # avoid problem report recursion
        except MessageParseError as e:
            self.logger.error(
                f"Message parsing failed: {str(e)}, sending problem report", exc_info=e
            )
            error_result = ProblemReport(
                description={
                    "en": str(e),
                    "code": "message-parse-failure",
                }
            )
            if inbound_message.receipt.thread_id:
                error_result.assign_thread_id(inbound_message.receipt.thread_id)

        trace_event(
            self.profile.settings,
            message,
            outcome="Dispatcher.handle_message.START",
        )

        context = RequestContext(profile)
        context.message = message
        context.message_receipt = inbound_message.receipt

        responder = DispatcherResponder(
            context,
            inbound_message,
            send_outbound,
            reply_session_id=inbound_message.session_id,
            reply_to_verkey=inbound_message.receipt.sender_verkey,
        )

        context.injector.bind_instance(BaseResponder, responder)

        # When processing oob attach message we supply the connection id
        # associated with the inbound message
        if inbound_message.connection_id:
            async with self.profile.session() as session:
                connection = await ConnRecord.retrieve_by_id(
                    session, inbound_message.connection_id
                )
        else:
            connection_mgr = BaseConnectionManager(profile)
            connection = await connection_mgr.find_inbound_connection(
                inbound_message.receipt
            )
            del connection_mgr

        if connection:
            inbound_message.connection_id = connection.connection_id

        context.connection_ready = connection and connection.is_ready
        context.connection_record = connection
        responder.connection_id = connection and connection.connection_id

        if error_result:
            await responder.send_reply(error_result)
        elif version_warning:
            await responder.send_reply(version_warning)
        elif context.message:
            context.injector.bind_instance(BaseResponder, responder)

            handler_cls = context.message.Handler
            handler = handler_cls().handle
            if self.collector:
                handler = self.collector.wrap_coro(handler, [handler.__qualname__])
            await handler(context, responder)

        trace_event(
            self.profile.settings,
            context.message,
            outcome="Dispatcher.handle_message.END",
            perf_counter=r_time,
        )

    async def make_message(self, profile: Profile, parsed_msg: dict) -> BaseMessage:
        """Deserialize a message dict into the appropriate message instance.

        Given a dict describing a message, this method
        returns an instance of the related message class.

        Args:
            parsed_msg: The parsed message
            profile: Profile

        Returns:
            An instance of the corresponding message class for this message

        Raises:
            MessageParseError: If the message doesn't specify @type
            MessageParseError: If there is no message class registered to handle
            the given type

        """
        if not isinstance(parsed_msg, dict):
            raise MessageParseError("Expected a JSON object")
        message_type = parsed_msg.get("@type")

        if not message_type:
            raise MessageParseError("Message does not contain '@type' parameter")

        if message_type.startswith("did:sov"):
            warnings.warn(
                "Received a core DIDComm protocol with the deprecated "
                "`did:sov:BzCbsNYhMrjHiqZDTUASHg;spec` prefix. The sending party should "
                "be notified that support for receiving such messages will be removed in "
                "a future release. Use https://didcomm.org/ instead.",
                DeprecationWarning,
            )
            self.logger.warning(
                "Received a core DIDComm protocol with the deprecated "
                "`did:sov:BzCbsNYhMrjHiqZDTUASHg;spec` prefix. The sending party should "
                "be notified that support for receiving such messages will be removed in "
                "a future release. Use https://didcomm.org/ instead.",
            )

        registry: ProtocolRegistry = self.profile.inject(ProtocolRegistry)
        try:
            message_cls = registry.resolve_message_class(message_type)
            if isinstance(message_cls, DeferLoad):
                message_cls = message_cls.resolved
        except ProtocolMinorVersionNotSupported as e:
            raise MessageParseError(f"Problem parsing message type. {e}")

        if not message_cls:
            raise MessageParseError(f"Unrecognized message type {message_type}")

        try:
            instance = message_cls.deserialize(parsed_msg)
        except BaseModelError as e:
            if "/problem-report" in message_type:
                raise ProblemReportParseError("Error parsing problem report message")
            raise MessageParseError(f"Error deserializing message: {e}") from e

        return instance

    async def complete(self, timeout: float = 0.1):
        """Wait for pending tasks to complete."""
        await self.task_queue.complete(timeout=timeout)


class DispatcherResponder(BaseResponder):
    """Handle outgoing messages from message handlers."""

    def __init__(
        self,
        context: RequestContext,
        inbound_message: InboundMessage,
        send_outbound: Coroutine,
        **kwargs,
    ):
        """Initialize an instance of `DispatcherResponder`.

        Args:
            context: The request context of the incoming message
            inbound_message: The inbound message triggering this handler
            send_outbound: Async function to send outbound message
            kwargs: Additional keyword arguments

        """
        super().__init__(**kwargs)
        # Weakly hold the context so it can be properly garbage collected.
        # Binding this DispatcherResponder into the context creates a circular
        # reference.
        self._context = weakref.ref(context)
        self._inbound_message = inbound_message
        self._send = send_outbound

    async def create_outbound(
        self, message: Union[AgentMessage, BaseMessage, str, bytes], **kwargs
    ) -> OutboundMessage:
        """Create an OutboundMessage from a message body.

        Args:
            message: The message payload
            kwargs: Additional keyword arguments

        Returns:
            OutboundMessage: The created outbound message.
        """
        context = self._context()
        if not context:
            raise RuntimeError("weakref to context has expired")

        if isinstance(message, AgentMessage) and context.settings.get("timing.enabled"):
            # Inject the timing decorator
            in_time = context.message_receipt and context.message_receipt.in_time
            if not message._decorators.get("timing"):
                message._decorators["timing"] = {
                    "in_time": in_time,
                    "out_time": datetime_now(),
                }

        return await super().create_outbound(message, **kwargs)

    async def send_outbound(
        self, message: OutboundMessage, **kwargs
    ) -> OutboundSendStatus:
        """Send outbound message.

        Args:
            message: The `OutboundMessage` to be sent
            kwargs: Additional keyword arguments
        """
        context = self._context()
        if not context:
            raise RuntimeError("weakref to context has expired")

        msg_type = kwargs.get("message_type")
        msg_id = kwargs.get("message_id")

        if (
            message.connection_id
            and msg_type
            and msg_type not in SKIP_ACTIVE_CONN_CHECK_MSG_TYPES
            and not await super().conn_rec_active_state_check(
                profile=context.profile,
                connection_id=message.connection_id,
            )
        ):
            raise RuntimeError(
                f"Connection {message.connection_id} is not ready"
                " which is required for sending outbound"
                f" message {msg_id} of type {msg_type}."
            )
        return await self._send(context.profile, message, self._inbound_message)

    async def send_webhook(self, topic: str, payload: dict):
        """Dispatch a webhook. DEPRECATED: use the event bus instead.

        Args:
            topic: the webhook topic identifier
            payload: the webhook payload value
        """
        warnings.warn(
            "responder.send_webhook is deprecated; please use the event bus instead.",
            DeprecationWarning,
        )
        context = self._context()
        if not context:
            raise RuntimeError("weakref to context has expired")

        await context.profile.notify("acapy::webhook::" + topic, payload)



"""aiohttp stats collector support."""

import aiohttp

from ..utils.stats import Collector


class StatsTracer(aiohttp.TraceConfig):
    """Attach hooks to client session events and report statistics."""

    def __init__(self, collector: Collector, prefix: str):
        """Initialize the `StatsTracer` instance."""
        super().__init__()
        self.collector = collector
        self.prefix = prefix
        self.on_request_start.append(self.request_start)
        self.on_connection_queued_start.append(self.connection_queued_start)
        self.on_connection_queued_end.append(self.connection_queued_end)
        self.on_dns_resolvehost_start.append(self.dns_resolvehost_start)
        self.on_dns_resolvehost_end.append(self.dns_resolvehost_end)
        self.on_connection_create_start.append(self.socket_connect_start)
        self.on_dns_cache_hit.append(self.socket_connect_start)  # restart timer
        self.on_dns_cache_miss.append(self.socket_connect_start)  # restart timer
        self.on_connection_reuseconn.append(self.connection_ready)
        self.on_connection_create_end.append(self.connection_ready)
        self.on_request_end.append(self.request_end)

    async def request_start(self, session, context, params):
        """Handle the start of a request."""
        context.method, context.url = params.method, params.url

    async def connection_queued_start(self, session, context, params):
        """Handle the start of a queued connection."""
        context.queue_timer = self.collector.timer(self.prefix + "queued").start()

    async def connection_queued_end(self, session, context, params):
        """Handle the end of a queued connection."""
        context.queue_timer.stop()

    async def dns_resolvehost_start(self, session, context, params):
        """Handle the start of a DNS resolution."""
        context.dns_timer = self.collector.timer(self.prefix + "dns_resolve").start()

    async def dns_resolvehost_end(self, session, context, params):
        """Handle the end of a DNS resolution."""
        context.dns_timer.stop()

    async def socket_connect_start(self, session, context, params):
        """Handle the start of a socket connection."""
        context.socket_timer = self.collector.timer(self.prefix + "connect").start()

    async def connection_ready(self, session, context, params):
        """Handle the end of connection acquisition."""
        try:
            context.socket_timer.stop()
        except AttributeError:
            pass
        context.fetch_timer = self.collector.timer(self.prefix + context.method).start()

    async def request_end(self, session, context, params):
        """Handle the end of request."""
        context.fetch_timer.stop()



"""Verifiable Credential and Presentation verification methods."""

import asyncio
from typing import List, Optional

from pyld.jsonld import JsonLdProcessor

from ..ld_proofs import (
    AuthenticationProofPurpose,
    CredentialIssuancePurpose,
    DocumentLoaderMethod,
    DocumentVerificationResult,
    LinkedDataProof,
    LinkedDataProofException,
    ProofPurpose,
)
from ..ld_proofs import verify as ld_proofs_verify
from .models.credential import VerifiableCredentialSchema
from .validation_result import PresentationVerificationResult


async def _verify_credential(
    *,
    credential: dict,
    suites: List[LinkedDataProof],
    document_loader: DocumentLoaderMethod,
    purpose: Optional[ProofPurpose] = None,
) -> DocumentVerificationResult:
    """Verify credential structure, proof purpose and signature."""

    # Validate credential structure
    errors = VerifiableCredentialSchema().validate(credential)
    if len(errors) > 0:
        raise LinkedDataProofException(
            f"Unable to verify credential with invalid structure: {errors}"
        )

    if not purpose:
        purpose = CredentialIssuancePurpose()

    result = await ld_proofs_verify(
        document=credential,
        suites=suites,
        purpose=purpose,
        document_loader=document_loader,
    )

    return result


async def verify_credential(
    *,
    credential: dict,
    suites: List[LinkedDataProof],
    document_loader: DocumentLoaderMethod,
    purpose: Optional[ProofPurpose] = None,
) -> DocumentVerificationResult:
    """Verify credential structure, proof purpose and signature.

    Args:
        credential (dict): The credential to verify
        suites (List[LinkedDataProof]): The signature suites to verify with
        document_loader (DocumentLoader): Document loader used for resolving of documents
        purpose (ProofPurpose, optional): Proof purpose to use.
            Defaults to CredentialIssuancePurpose

    Returns:
        DocumentVerificationResult: The result of the verification. Verified property
            indicates whether the verification was successful

    """
    try:
        return await _verify_credential(
            credential=credential,
            document_loader=document_loader,
            suites=suites,
            purpose=purpose,
        )
    except Exception as e:
        return DocumentVerificationResult(verified=False, document=credential, errors=[e])


async def _verify_presentation(
    *,
    presentation: dict,
    suites: List[LinkedDataProof],
    document_loader: DocumentLoaderMethod,
    challenge: Optional[str] = None,
    domain: Optional[str] = None,
    purpose: Optional[ProofPurpose] = None,
):
    """Verify presentation structure, credentials, proof purpose and signature."""

    if not purpose and not challenge:
        raise LinkedDataProofException(
            'A "challenge" param is required for AuthenticationProofPurpose.'
        )
    elif not purpose:
        purpose = AuthenticationProofPurpose(challenge=challenge, domain=domain)

    # TODO validate presentation structure here
    if "proof" not in presentation:
        raise LinkedDataProofException('presentation must contain "proof"')

    presentation_result = await ld_proofs_verify(
        document=presentation,
        suites=suites,
        purpose=purpose,
        document_loader=document_loader,
    )

    credential_results = None

    credentials = JsonLdProcessor.get_values(presentation, "verifiableCredential")
    credential_results = await asyncio.gather(
        *[
            verify_credential(
                credential=credential,
                suites=suites,
                document_loader=document_loader,
                # FIXME: we don't want to inherit the authentication purpose
                # from the presentation. However we do want to have subject
                # authentication I guess
                # purpose=purpose,
            )
            for credential in credentials
        ]
    )

    credentials_verified = all(result.verified for result in credential_results)
    verified = credentials_verified and presentation_result.verified

    return PresentationVerificationResult(
        verified=verified,
        presentation_result=presentation_result,
        credential_results=credential_results,
        errors=presentation_result.errors,
    )


async def verify_presentation(
    *,
    presentation: dict,
    suites: List[LinkedDataProof],
    document_loader: DocumentLoaderMethod,
    purpose: Optional[ProofPurpose] = None,
    challenge: Optional[str] = None,
    domain: Optional[str] = None,
) -> PresentationVerificationResult:
    """Verify presentation structure, credentials, proof purpose and signature.

    Args:
        presentation (dict): The presentation to verify
        suites (List[LinkedDataProof]): The signature suites to verify with
        document_loader (DocumentLoader): Document loader used for resolving of documents
        purpose (ProofPurpose, optional): Proof purpose to use.
            Defaults to AuthenticationProofPurpose
        challenge (str, optional): The challenge to use for authentication.
            Required if purpose is not passed, not used if purpose is passed
        domain (str, optional): Domain to use for the authentication proof purpose.
            Not used if purpose is passed

    Returns:
        PresentationVerificationResult: The result of the verification. Verified property
            indicates whether the verification was successful

    """

    # TODO: I think we should add some sort of options to authenticate the subject id
    # to the presentation verification method controller

    try:
        return await _verify_presentation(
            presentation=presentation,
            challenge=challenge,
            purpose=purpose,
            suites=suites,
            domain=domain,
            document_loader=document_loader,
        )
    except Exception as e:
        return PresentationVerificationResult(verified=False, errors=[e])


__all__ = ["verify_presentation", "verify_credential"]



"""DIF Proof Schema."""

from marshmallow import fields

from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import UUID4_EXAMPLE, UUID4_VALIDATE, StrOrDictField
from ....vc.vc_ld import LinkedDataProofSchema
from .pres_exch import PresentationSubmissionSchema


class DIFProofSchema(OpenAPISchema):
    """Schema for DIF Proof."""

    id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={"description": "ID", "example": UUID4_EXAMPLE},
    )
    contexts = fields.List(StrOrDictField(), data_key="@context", required=True)
    types = fields.List(
        fields.Str(metadata={"description": "Types"}), data_key="type", required=True
    )
    credentials = fields.List(
        fields.Dict(required=False, metadata={"description": "Credentials"}),
        data_key="verifiableCredential",
    )
    proof = fields.Nested(
        LinkedDataProofSchema(),
        required=False,
        metadata={"description": "The proof of the credential"},
    )
    presentation_submission = fields.Nested(PresentationSubmissionSchema())
